# Copyright 2016 Matthew J. Aburn
# 
# This program is free software: you can redistribute it and/or modify 
# it under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License, or 
# (at your option) any later version. See <http://www.gnu.org/licenses/>.

"""
Implements the core functionality of nsim.

Classes:
--------
``Model``   base class for different kinds of dynamical model
``ODEModel``   system of ordinary differential equations
``ItoModel``   system of Ito stochastic differential equations
``StratonovichModel``  system of Stratonovich stochastic differential equations
``NetworkModel``   many coupled instances of a submodel connected in a network

``Simulation``   single simulation run of a model, with simulation results
``RepeatedSim``   repeated simulations of the same model (to get statistics)
``ParameterSim``  multiple simulations of a model exploring parameter space

``RemoteTimeseries`` Local proxy representing a Timeseries on a remote engine 
``DistTimeseries`` Timeseries with one axis distributed onto multiple engines 

functions:
----------
``newmodel()``  Create a new model class dynamically at runtime
``newsim()``  Create a new simulation dynamically at runtime
"""

from __future__ import absolute_import
from .timeseries import Timeseries, _Timeslice
from . import analysesN
import sdeint
import distob
from scipy import stats
from scipy import integrate
import numpy as np
from collections import Sequence
import copy
import types
import warnings
import numbers
import random
#from memory_profiler import profile

# types for compatibility across python 2 and 3
_SliceType = type(slice(None))
_EllipsisType = type(Ellipsis)
_TupleType = type(())
_NewaxisType = type(np.newaxis)


class Error(Exception):
    pass


class SimTypeError(Error):
    pass


class SimValueError(Error):
    pass


@distob.proxy_methods(Timeseries, exclude=('dtype',), include_underscore=(
    '__getitem__', '__setitem__', '__getslice__', '__setslice__'))
class RemoteTimeseries(distob.RemoteArray, object):
    """Local object representing a Timeseries that may be local or remote"""

    def __repr__(self):
        return distob.methodcall(self, '__repr__').replace(
            self._ref.type.__name__, self.__class__.__name__, 1)

    def __numpy_ufunc__(self, ufunc, method, i, inputs, **kwargs):
        out_arr = super(RemoteTimeseries, self).__numpy_ufunc__(
                ufunc, method, i, inputs, **kwargs)
        return _ufunc_wrap(out_arr, ufunc, method, i, inputs, **kwargs)

    __array_priority__ = 3.0

    def __array_prepare__(self, out_arr, context=None):
        """Fetch underlying data to user's computer and apply ufunc locally.
        Only used as a fallback, for numpy versions < 1.10.0 which lack 
        support for the __numpy_ufunc__ mechanism. 
        """
        #print('RemoteTimeseries __array_prepare__ context=%s' % repr(context))
        out_arr = super(RemoteTimeseries, self).__array_prepare__(
                out_arr, context)
        if context is None:
            return out_arr
        else:
            ufunc, inputs, i = context
            return _ufunc_wrap(out_arr, ufunc, None, i, inputs)

    def plot(self, title=None, show=True):
        self._fetch()
        return self._obcache.plot(title, show)

    def psd(self, plot=True):
        """Power spectral density
        Args:
          plot (bool)
        """
        if not plot:
            return distob.methodcall(self, 'psd', plot=False)
        else:
            self._fetch()
            return self._obcache.psd(plot=True)

    def concatenate(self, tup, axis=0):
        if not isinstance(tup, Sequence):
            tup = (tup,)
        if tup is (None,) or len(tup) is 0:
            return self
        baretup = (self.view(np.ndarray),) + tuple(tup)
        tup = (self,) + tuple(tup)
        new_array = distob.concatenate(baretup, axis)
        if not all(hasattr(ts, 'tspan') and
                   hasattr(ts, 'labels') for ts in tup):
            return new_array
        if axis == 0:
            starts = [ts.tspan[0] for ts in tup]
            ends = [ts.tspan[-1] for ts in tup]
            if not all(starts[i] > ends[i-1] for i in range(1, len(starts))):
                # series being joined are not ordered in time. not a Timeseries
                return new_array
            else:
                new_tspan = np.concatenate([ts.tspan for ts in tup])
        else:
            new_tspan = self.tspan
        new_labels = [None]
        for ax in range(1, new_array.ndim):
            if ax == axis:
                axislabels = []
                for ts in tup:
                    if ts.labels[axis] is None:
                        axislabels.extend('' * ts.shape[axis])
                    else:
                        axislabels.extend(ts.labels[axis])
                if all(lab == '' for lab in axislabels):
                    new_labels.append(None)
                else:
                    new_labels.append(axislabels)
            else:
                # non-concatenation axis
                axlabels = tup[0].labels[ax]
                if not all(ts.labels[ax] == axlabels for ts in tup[1:]):
                    # series to be joined do not agree on labels for this axis
                    axlabels = None
                new_labels.append(axlabels)
        if isinstance(new_array, distob.RemoteArray):
            # TODO: could have kept tspan and labels remote in this case
            return _rts_from_ra(new_array, new_tspan, new_labels)
        else:
            assert(isinstance(new_array, distob.DistArray))
            return _dts_from_da(new_array, new_tspan, new_labels)


@distob.proxy_methods(_Timeslice, include_underscore=(
    '__getitem__', '__setitem__', '__getslice__', '__setslice__', '__repr__'))
class Remote_Timeslice(distob.Remote, object):
    pass


class DistTimeseries(distob.DistArray):
    """a Timeseries with one axis distributed onto multiple computing engines.

    For example, a multi-channel timeseries could be distributed so that each
    channel is held on a different computer (for parallel computation)

    Currently only a non-time axis can be distributed.
    """
    # TODO: consider refactoring as __getitem__ and __repr__ share much logic
    #   with the Timeseries class
    def __init__(self, subts, axis=None, axislabels=None):
        """Make a DistTimeseries from a list of existing remote Timeseries

        Args:
          subts (list of RemoteTimeseries, or list of Ref to Timeseries):
            the sub-timeseries (possibly remote) which form the whole
            DistTimeseries when concatenated. These must all have the same
            time points, dtype and must have the same shape excepting the
            distributed axis.

          axis (int, optional): Position of the distributed axis, which is the
            axis along which the sub-timeseries will be concatenated. Default
            is the last axis. Cannot be 0, that is, we currently only allow a
            non-time axis to be distributed.

          axislabels (list of str, optional): names to label each position 
            along the new distributed axis. e.g. ['node1', ..., 'nodeN']
        """
        if axis == 0:
            raise SimValueError(u'Currently cannot distribute the time axis')
        if axislabels is not None and not isinstance(axislabels, Sequence):
            raise SimValueError(u'axislabels should be a list of str')
        dist_length = sum(ts.shape[axis] for ts in subts)
        if (axislabels is not None and len(axislabels) != dist_length):
            raise SimValueError(
                    u'mismatch: got %d axislabels for axis %d of length %d' % (
                        len(axislabels), axis, dist_length))
        super(DistTimeseries, self).__init__(subts, axis)
        self.tspan = distob.gather(self._subarrays[0].tspan)
        # Expensive to validate all tspans are the same. check start and end t
        # (TODO: in Timeseries class make special case for constant timestep)
        starts = [rts.tspan[0] for rts in self._subarrays]
        ends = [rts.tspan[-1] for rts in self._subarrays]
        if (not all(t == self.tspan[0] for t in starts) or 
                not all(t == self.tspan[-1] for t in ends)):
            raise SimValueError(u'timeseries must use the same time points')
        nlabels = [rts.labels for rts in self._subarrays]
        self.labels = [None] * self.ndim
        # only give labels to an axis if all sub-timeseries agree
        for i in range(1, self.ndim):
            if i is self._distaxis:
                self.labels[i] = axislabels
            else:
                candidate = nlabels[0][i]
                if all(labels[i] == candidate for labels in nlabels[1:]):
                    self.labels[i] = candidate
        if self._obcache_current:
            self._obcache.labels[self._distaxis] = axislabels
        self.t = _Timeslice(self)

    def _fetch(self):
        """forces update of a local cached copy of the real object
        (regardless of the preference setting self.cache)"""
        if not self._obcache_current:
            from distob import engine
            ax = self._distaxis
            self._obcache = distob.concatenate(
                    [ra._ob for ra in self._subarrays], ax)
            if hasattr(self, 'labels'):
                self._obcache.labels[ax] = self.labels[ax]
            # let subarray obcaches and main obcache be views on same memory:
            for i in range(self._n):
                ix = [slice(None)] * self.ndim
                ix[ax] = slice(self._si[i], self._si[i+1])
                self._subarrays[i]._obcache = self._obcache[tuple(ix)]
            self._obcache_current = True
            # now prefer local processing:
            self.__engine_affinity__ = (
                    engine.eid, self.__engine_affinity__[1])

    def __ob(self):
        """return a copy of the real object"""
        self._fetch()
        return self._obcache

    _ob = property(fget=__ob, doc='return a local copy of the object')

    @classmethod
    def add_analyses(cls, source, vectorize=False):
        """Dynamically add new analysis methods to the DistTimeseries class.
        Args:
          source: Can be a function, module or the filename of a python file.
            If a filename or a module is given, then all functions defined 
            inside not starting with _ will be added as methods.
          vectorize (bool): Whether to apply `distob.vectorize()` to the 
            function before making it a method of DistTimeseries.

        The only restriction on the functions is that they can accept a 
        Timeseries as their first argument.
        """
        if isinstance(source, types.FunctionType):
            if vectorize:
                source = distob.vectorize(source)
            setattr(cls, source.__name__, source)
        else:
            if isinstance(source, types.ModuleType):
                mod = source
            elif isinstance(source, types.StringTypes):
                import os
                import imp
                path = os.path.abspath(source)
                if os.path.isfile(path) and path[-3:] == '.py':
                    dir, file = os.path.split(path)
                    name = file[:-3]
                    module_info = imp.find_module(name, [dir])
                    mod = imp.load_module('nsim.' + name, *module_info)
                elif (os.path.isdir(path) and 
                        '__init__.py' in os.listdir(path)):
                    module_info = imp.find_module('__init__', [path])
                    name = os.path.basename(path)
                    mod = imp.load_module('nsim.' + name, *module_info)
                else:
                    raise Error('"%s" is not a file or directory' % source)
            else:
                raise ValueError('`source` argument not a function or module')
            for name, obj in mod.__dict__.items():
                if name[0] != '_' and isinstance(obj, types.FunctionType):
                    if vectorize:
                        obj = distob.vectorize(obj)
                    setattr(cls, name, obj)

    def __getitem__(self, index):
        """Slice the distributed timeseries"""
        ar = super(DistTimeseries, self).__getitem__(index)
        if isinstance(ar, distob.RemoteArray) or isinstance(ar, np.ndarray):
            # slicing result is no longer distributed
            return ar
        # otherwise `ar` is a DistArray
        if (not isinstance(ar._subarrays[0], RemoteTimeseries) and
                not isinstance(ar._subarrays[0], Timeseries)):
            return ar
        cur_labels = self.labels
        cur_shape = self.shape # current values, may be updated by np.newaxis
        cur_ndim = self.ndim
        if isinstance(index, np.ndarray) and index.dtype.type is np.bool_:
            raise Error('indexing by boolean array not yet implemented')
        if not isinstance(index, Sequence):
            index = (index,) + (slice(None),)*(self.ndim - 1)
        ix_types = tuple(type(x) for x in index)
        if (np.ndarray in ix_types or
                (not isinstance(index, _TupleType) and
                    _NewaxisType not in ix_types and
                    _EllipsisType not in ix_types and
                    _SliceType not in ix_types) or
                any(issubclass(T, Sequence) for T in ix_types)):
            basic_slicing = False
        else:
            basic_slicing = True
        # Apply any ellipses
        while _EllipsisType in ix_types:
            pos = ix_types.index(_EllipsisType)
            m = (self.ndim + ix_types.count(_NewaxisType) - len(index) + 1)
            index = index[:pos] + (slice(None),)*m + index[(pos+1):]
            ix_types = tuple(type(x) for x in index)
        # Apply any np.newaxis
        while _NewaxisType in ix_types:
            pos = ix_types.index(type(np.newaxis))
            if pos is 0:
                # prepended an axis: no longer a Timeseries
                return ar
            index = index[:pos] + (slice(None),) + index[(pos+1):]
            cur_labels = cur_labels[:pos] + [None] + cur_labels[pos:]
            cur_shape = cur_shape[:pos] + (1,) + cur_shape[pos:]
            cur_ndim = len(cur_shape)
            ix_types = tuple(type(x) for x in index)
        index = tuple(index) + (slice(None),)*(cur_ndim - len(index))
        if len(index) > cur_ndim:
            raise IndexError('too many indices for array')
        if basic_slicing:
            new_tspan = self.tspan[index[0]]
            if (not isinstance(new_tspan, np.ndarray) or 
                    new_tspan.shape is () or
                    new_tspan.shape[0] is 0):
                # axis 0 has been sliced away: result is no longer a Timeseries
                return ar
            new_labels = [None]
            for i in range(1, cur_ndim):
                # make temp ndarray of labels to ensure correct slicing
                if cur_labels[i] is None:
                    labelarray = np.array([''] * cur_shape[i])
                else:
                    labelarray = np.array(cur_labels[i])
                newlabelarray = labelarray[index[i]]
                if newlabelarray.shape is not ():
                    if cur_labels[i] is None:
                        new_labels.append(None)
                    else:
                        new_labels.append(list(newlabelarray))
            new_subts = ar._subarrays
            new_distaxis = ar._distaxis
            new_axislabels = new_labels[ar._distaxis]
            return DistTimeseries(new_subts, new_distaxis, new_axislabels)
        else:
            # advanced integer indexing
            is_fancy = tuple(not isinstance(x, _SliceType) for x in index)
            fancy_pos = tuple(i for i in range(len(index)) if is_fancy[i])
            nonfancy_pos = tuple(
                    i for i in range(len(index)) if not is_fancy[i])
            contiguous = (fancy_pos[-1] - fancy_pos[0] == len(fancy_pos) - 1)
            index = list(index)
            ix_arrays = [index[i] for i in fancy_pos]
            ix_arrays = np.broadcast_arrays(*ix_arrays)
            for j in range(len(fancy_pos)):
                if ix_arrays[j].shape is ():
                    ix_arrays[j] = np.expand_dims(ix_arrays[j], 0)
                index[fancy_pos[j]] = ix_arrays[j]
            index = tuple(index)
            ishape = index[fancy_pos[0]].shape # common shape all index arrays
            idim = len(ishape)
            assert(idim > 0)
            new_tspan = None
            if contiguous and not is_fancy[0]:
                new_tspan = self.tspan[index[0]]
                if (not isinstance(new_tspan, np.ndarray) or
                        new_tspan.shape is () or
                        new_tspan.shape[0] is 0):
                    # axis 0 has been sliced away: no longer a Timeseries
                    return ar
            # compute labels for the nonfancy output axes
            nonfancy_labels = []
            nonfancy_retained = []
            for i in nonfancy_pos:
                # make temp ndarray of labels to ensure correct slicing
                if cur_labels[i] is None:
                    labelarray = np.array([''] * cur_shape[i])
                else:
                    labelarray = np.array(cur_labels[i])
                newlabelarray = labelarray[index[i]]
                if newlabelarray.shape is ():
                    nonfancy_retained.append(False)
                else:
                    nonfancy_retained.append(True)
                    if cur_labels[i] is None or len(newlabelarray) is 0:
                        nonfancy_labels.append(None)
                    else:
                        nonfancy_labels.append(list(newlabelarray))
            # compute labels for the fancy output axes:
            #
            # For each fancy output axis k, call input axis i a 'candidate'
            # label source for k if k is the only nonconstant axis in the
            # indexing array for i.
            # We will give labels/tspan to output axis k from input axis i
            # only if i is the sole candidate label source for k.
            candidates = [[]] * idim
            # candidates[k] will be a list of candidate label sources for k
            for i in fancy_pos:
                nonconstant_ix_axes = []
                for k in range(idim):
                    n = ishape[k]
                    if n > 0:
                        partix = np.split(index[i], n, axis=k)
                        if not all(np.array_equal(
                                partix[0], partix[q]) for q in range(1, n)):
                            nonconstant_ix_axes.append(k)
                if len(nonconstant_ix_axes) is 1:
                    candidates[nonconstant_ix_axes[0]].append(i)
            fancy_labels = []
            for k in range(idim):
                if len(candidates[k]) is 1:
                    # then we can label this output axis
                    label_source = candidates[k][0]
                    if cur_labels[label_source] is None:
                        labelarray = np.array([''] * cur_shape[label_source])
                    else:
                        labelarray = np.array(cur_labels[label_source])
                    iix = [0] * idim
                    iix[k] = slice(None)
                    iix = tuple(iix)
                    newlabelarray = labelarray[index[label_source][iix]]
                    if newlabelarray.shape is not ():
                        if cur_labels[label_source] is None:
                            fancy_labels.append(None)
                        else:
                            fancy_labels.append(list(newlabelarray))
                    if k is 0 and (is_fancy[0] or not contiguous):
                        # then this output axis will be axis 0 of output
                        if label_source is 0:
                            new_tspan = self.tspan[index[0][iix]]
                            if (not isinstance(new_tspan, np.ndarray) or
                                    new_tspan.shape is () or
                                    new_tspan.shape[0] is 0):
                                # axis 0 has been sliced away: not a Timeseries
                                return ar
                            if not np.all(np.diff(new_tspan) > 0):
                                #tspan not monotonic increasing: not Timeseries
                                return ar
                        else:
                            #axis 0 no longer represents time: not a Timeseries
                            return ar
                else:
                    # not a 'sole candidate'
                    fancy_labels.append(None)
            if contiguous:
                # fancy output axes are put where the fancy input axes were:
                new_labels = []
                for i in range(0, fancy_pos[0]):
                    if nonfancy_retained.pop(0):
                        new_labels.append(nonfancy_labels.pop(0))
                new_labels.extend(fancy_labels)
                for i in range(fancy_pos[-1] + 1, cur_ndim):
                    if nonfancy_retained.pop(0):
                        new_labels.append(nonfancy_labels.pop(0))
            else:
                # not contiguous. fancy output axes move to the start:
                new_labels = fancy_labels + nonfancy_labels
            if new_tspan is None:
                return ar
            else:
                new_subts = ar._subarrays
                new_distaxis = ar._distaxis
                new_axislabels = new_labels[ar._distaxis]
                return DistTimeseries(new_subts, new_distaxis, new_axislabels)

    def __repr__(self):
        classname = self.__class__.__name__
        super_classname = super(DistTimeseries, self).__class__.__name__
        if self.tspan.shape is ():
            first = last = self.tspan
        else:
            first = self.tspan[0]
            last = self.tspan[-1]
        head = (u'<%s of shape %s from time %f to %f '
                 'with axis %d distributed>:\n') % (
                     classname, self.shape, first, last, self._distaxis)
        repr_tspan = 'tspan=' + repr(self.tspan)
        if len(repr_tspan) > 160:
            repr_tspan = 'tspan=array([ %f, ..., %f ])' % (first, last)
        # reuse DistArray repr, but replacing the heading:
        content = super(DistTimeseries, self).__repr__()
        first_newline = content.index('\n')
        content = content[(first_newline + 1):]
        content = content.replace(super_classname, classname, 1)
        content = content.rstrip(')') + ', \n' + repr_tspan
        if all(l is None for l in self.labels):
            labelsrep = ''
        else:
            labelsrep = ', \nlabels=['
            for lab in self.labels:
                if len(repr(lab)) > 160:
                    labelsrep += '[%s, ..., %s], ' % (lab[0], lab[-1])
                else:
                    labelsrep += repr(lab) + ', '
            labelsrep = labelsrep[:-2] + ']'
        return head + content + labelsrep + ')'

    def __numpy_ufunc__(self, ufunc, method, i, inputs, **kwargs):
        out_arr = super(DistTimeseries, self).__numpy_ufunc__(
                ufunc, method, i, inputs, **kwargs)
        return _ufunc_wrap(out_arr, ufunc, method, i, inputs, **kwargs)

    __array_priority__ = 3.0

    def __array_prepare__(self, out_arr, context=None):
        """Fetch underlying data to user's computer and apply ufunc locally.
        Only used as a fallback, for numpy versions < 1.10.0 which lack 
        support for the __numpy_ufunc__ mechanism. 
        """
        #print('DistTimeseries __array_prepare__ context=%s' % repr(context))
        out_arr = super(DistTimeseries, self).__array_prepare__(
                out_arr, context)
        if context is None:
            return out_arr
        else:
            ufunc, inputs, i = context
            return _ufunc_wrap(out_arr, ufunc, None, i, inputs)

    @classmethod
    def __distob_vectorize__(cls, f):
        """Upgrades a normal function f to act on a DistTimeseries in parallel

        Args:
          f (callable): ordinary function which expects as its first
            argument a Timeseries (of the same shape as our subarrays)

        Returns:
          vf (callable): new function that takes a DistTimeseries as its first
            argument. ``vf(dist_timeseries)`` will do the computation ``f(ts)``
            on each sub-timeseries in parallel and if possible will return the
            results as a DistTimeseries or DistArray. (or if the results are
            not arrays, will return a list with the result for each
            sub-timeseries)
        """
        # TODO: shares much code with the superclass method. refactor.
        def vf(self, *args, **kwargs):
            kwargs = kwargs.copy()
            kwargs['block'] = False
            kwargs['prefer_local'] = False
            ars = [distob.call(
                    f, ra, *args, **kwargs) for ra in self._subarrays]
            results = [distob.convert_result(ar) for ar in ars]
            if all(isinstance(r, distob.RemoteArray) and
                    r.ndim == results[0].ndim for r in results):
                # Then we will join the results and return a DistArray.
                out_shapes = [ra.shape for ra in results]
                new_distaxis = self._new_distaxis(out_shapes)
                if new_distaxis is None:
                    return results # incompatible array shapes. return a list.
                if new_distaxis == results[0].ndim:
                    results = [r.expand_dims(new_distaxis) for r in results]
                if all(isinstance(r, RemoteTimeseries) for r in results):
                    if (sum(r.shape[new_distaxis] for r in results) ==
                            self.shape[self._distaxis]):
                        axlabels = self.labels[self._distaxis]
                    else:
                        axlabels = None
                    try:
                       return DistTimeseries(results, new_distaxis, axlabels)
                    except SimValueError:
                        pass
                return distob.DistArray(results, new_distaxis)
            elif all(isinstance(r, numbers.Number) for r in results):
                return np.array(results)
            else:
                return results  # not arrays or numbers. return a list
        if hasattr(f, '__name__'):
            vf.__name__ = 'v' + f.__name__
            f_str = f.__name__ + '()'
        else:
            f_str = 'callable'
        doc = u"""Apply %s in parallel to a DistTimeseries\n
               Args:
                 dts (DistTimeseries)
                 other args are the same as for %s
               """ % (f_str, f_str)
        if hasattr(f, '__doc__') and f.__doc__ is not None:
            doc = doc.rstrip() + (' detailed below:\n----------\n' + f.__doc__)
        vf.__doc__ = doc
        return vf

    def expand_dims(self, axis):
        """Insert a new axis, at a given position in the array shape
        Args:
          axis (int): Position (amongst axes) where new axis is to be inserted.
        """
        if axis <= self._distaxis:
            subaxis = axis
            new_distaxis = self._distaxis + 1
        else:
            subaxis = axis - 1
            new_distaxis = self._distaxis
        new_subts = [rts.expand_dims(subaxis) for rts in self._subarrays]
        if axis == 0:
            # prepended an axis: no longer a Timeseries
            return distob.DistArray(new_subts, new_distaxis)
        else:
            axislabels = self.labels[self._distaxis]
            return DistTimeseries(new_subts, new_distaxis, axislabels)

    def absolute(self):
        """Calculate the absolute value element-wise.

        Returns:
          absolute (Timeseries):
            Absolute value. For complex input (a + b*j) gives sqrt(a**a + b**2)
        """
        da = distob.vectorize(np.absolute)(self)
        return _dts_from_da(da, self.tspan, self.labels)

    def abs(self):
        """Calculate the absolute value element-wise."""
        return self.absolute()

    def angle(self, deg=False):
        """Return the angle of a complex Timeseries

        Args:
          deg (bool, optional):
            Return angle in degrees if True, radians if False (default).

        Returns:
          angle (Timeseries):
            The counterclockwise angle from the positive real axis on
            the complex plane, with dtype as numpy.float64.
        """
        if self.dtype.str[1] != 'c':
            warnings.warn('angle() is intended for complex-valued timeseries',
                          RuntimeWarning, 1)
        da = distob.vectorize(np.angle)(self, deg)
        return _dts_from_da(da, self.tspan, self.labels)


def _ufunc_wrap(out_arr, ufunc, method, i, inputs, **kwargs):
    """After using the superclass __numpy_ufunc__ to route ufunc computations 
    on the array data, convert any resulting ndarray, RemoteArray and DistArray
    instances into Timeseries, RemoteTimeseries and DistTimeseries instances
    if appropriate"""
    # Assigns tspan/labels to an axis only if inputs do not disagree on them.
    shape = out_arr.shape
    ndim = out_arr.ndim
    if ndim is 0 or shape[0] is 0:
        # not a timeseries
        return out_arr
    candidates = [a.tspan for a in inputs if (hasattr(a, 'tspan') and
                                              a.shape[0] == shape[0])]
    # Expensive to validate all tspans are the same. check start and end t
    starts = [tspan[0] for tspan in candidates]
    ends = [tspan[-1] for tspan in candidates]
    if len(set(starts)) != 1 or len(set(ends)) != 1:
        # inputs cannot agree on tspan
        return out_arr
    else:
        new_tspan = candidates[0]
    new_labels = [None]
    for i in range(1, ndim):
        candidates = [a.labels[i] for a in inputs if (hasattr(a, 'labels') and 
                 a.shape[i] == shape[i] and a.labels[i] is not None)] 
        if len(candidates) is 1:
            new_labels.append(candidates[0])
        elif (len(candidates) > 1 and all(labs[j] == candidates[0][j] for 
                labs in candidates[1:] for j in range(shape[i]))):
            new_labels.append(candidates[0])
        else:
            new_labels.append(None)
    if isinstance(out_arr, np.ndarray):
        return Timeseries(out_arr, new_tspan, new_labels)
    elif isinstance(out_arr, distob.RemoteArray):
        return _rts_from_ra(out_arr, new_tspan, new_labels)
    elif (isinstance(out_arr, distob.DistArray) and
          all(isinstance(ra, RemoteTimeseries) for ra in out_arr._subarrays)):
        return _dts_from_da(out_arr, new_tspan, new_labels)
    else:
        return out_arr


def _rts_from_ra(ra, tspan, labels, block=True):
    """construct a RemoteTimeseries from a RemoteArray"""
    def _convert(a, tspan, labels):
        from nsim import Timeseries
        return Timeseries(a, tspan, labels)
    return distob.call(
            _convert, ra, tspan, labels, prefer_local=False, block=block)


def _dts_from_da(da, tspan, labels):
    """construct a DistTimeseries from a DistArray"""
    sublabels = labels[:]
    new_subarrays = []
    for i, ra in enumerate(da._subarrays):
        if isinstance(ra, RemoteTimeseries):
            new_subarrays.append(ra)
        else:
            if labels[da._distaxis]:
                sublabels[da._distaxis] = labels[da._distaxis][
                        da._si[i]:da._si[i+1]]
            new_subarrays.append(_rts_from_ra(ra, tspan, sublabels, False))
    new_subarrays = [distob.convert_result(ar) for ar in new_subarrays]
    da._subarrays = new_subarrays
    da.__class__ = DistTimeseries
    da.tspan = tspan
    da.labels = labels
    da.t = _Timeslice(da)
    return da


class Model(object):
    """Base class for different kinds of dynamical systems

    Attributes:
      integrator (sequence containing a single function): Which function to use
        by default to integrate systems of this class.
    """
    integrator = (None,)

    def __init__(self):
        """When making each new instance from the Model class, we will convert
        any random-variable class attribute into a fixed number drawn from the
        specified distribution. Thus each individual object made from the 
        class 'recipe' will receive slightly different parameter values.
        """
        for attrib in dir(self):
            if isinstance(getattr(self,attrib), stats.distributions.rv_frozen):
                setattr(self, attrib, getattr(self,attrib).rvs())

    def integrate(self, tspan):
        pass 


class _DEModel(Model):
    """Base class with some common code shared by the finite dimensional DE
    systems ODEModel, ItoModel and StratonovichModel

    Attributes:
      dimension (integer): Dimension of the state space

    output_vars (list of integers): If i is in this list then y[i] is 
      considered an output variable

    y0 (array of shape (dimension,)): Initial state

    labels (sequence of str): optional names for the dynamical variables

    integrator (sequence containing a single function): Which function to use
      by default to integrate systems of this class
    """
    y0 = np.array([0.0])
    output_vars = [0]
    labels = None
    integrator = (None,)

    def __init__(self):
        super(_DEModel, self).__init__()
        # do some validation of attributes
        if not hasattr(self, 'y0'):
            raise Exception(
                    """%s instance should have a `y0` attribute to define the
                    initial state""" % self.__class__.__name__)
        y0_length = (1 if isinstance(self.y0, numbers.Number) else
                     np.asarray(self.y0).shape[0])
        if not hasattr(self, 'dimension'):
            self.dimension = y0_length
        if self.dimension != y0_length:
            raise Exception(
                    """For a system of dimension %d, initial state y0 should
                    have length %d.""" % (self.dimension, self.dimension))
        # by default consider all variables to be outputs
        if not hasattr(self, 'output_vars'):
            self.output_vars = range(self.dimension)
        if (self.labels is not None and
                len(self.labels) != self.dimension):
            raise Excpetion(
                    """System has dimension %d. If labels are provided to name
                    the dynamical variables there should be %d labels but got
                    %d.""" %(self.dimension, self.dimension, len(self.labels)))


class ODEModel(_DEModel):
    """Model defined by a system of ordinary differential equations
    dy/dt = f(y, t)

    Attributes:
      dimension (integer): Dimension of the state space

      output_vars (list of integers): If i is in this list then y[i] is 
        considered an output variable

      y0 (array of shape (dimension,)): Initial state

      labels (sequence of str): optional names for the dynamical variables

      integrator (sequence containing a single function): Which function to use
        by default to integrate systems of this class

    Methods:
      f(y, t): right hand side of the ODE system
    """
    y0 = np.array([0.0])
    output_vars = [0]
    labels = None
    integrator = (integrate.odeint,)

    def __init__(self):
        """Create an instance of this system, ready to simulate"""
        super(ODEModel, self).__init__()

    def integrate(self, tspan):
        ar = self.integrator[0](self.f, self.y0, tspan)
        return Timeseries(ar, tspan)

    def f(self, y, t):
        pass


class ItoModel(_DEModel):
    """Model defined by system of Ito stochastic differential equations
    dy = f(y, t) dt + G(y, t) dW

    Attributes:
      dimension (integer): Dimension of the state space

      output_vars (list of integers): If i is in this list then y[i] is 
        considered an output variable

      y0 (array of shape (dimension,)): Initial state

      labels (sequence of str): optional names for the dynamical variables

      integrator (sequence containing a single function): Which function to use
        by default to integrate systems of this class.

    Methods:
      f(y, t): deterministic part of Ito SDE system 
      G(y, t): noise coefficient matrix of Ito SDE system 
    """
    y0 = np.array([0.0])
    output_vars = [0]
    labels = None
    integrator = (sdeint.itoint,)

    def __init__(self):
        """Create an instance of this system, ready to simulate"""
        super(ItoModel, self).__init__()

    def integrate(self, tspan):
        ar = self.integrator[0](self.f, self.G, self.y0, tspan)
        return Timeseries(ar, tspan)

    def f(self, y, t):
        pass

    def G(self, y, t):
        pass


class StratonovichModel(_DEModel):
    """Model defined by system of Stratonovich stochastic differential
    equations   dy = f(y, t) dt + G(y, t) \circ dW

    Attributes:
      dimension (integer): Dimension of the state space

      output_vars (list of integers): If i is in this list then y[i] is 
        considered an output variable

      y0 (array of shape (dimension,)): Initial state

      labels (sequence of str): optional names for the dynamical variables

      integrator (sequence containing a single function): Which function to use
        by default to integrate systems of this class.

    Methods:
      f(y, t): deterministic part of Stratonovich SDE system 
      G(y, t): noise coefficient matrix of Stratonovich SDE system 
    """
    y0 = np.array([0.0])
    output_vars = [0]
    labels = None
    integrator = (sdeint.stratint,)

    def __init__(self):
        """Create an instance of this system, ready to simulate"""
        super(StratonovichModel, self).__init__()

    def integrate(self, tspan):
        ar = self.integrator[0](self.f, self.G, self.y0, tspan)
        return Timeseries(ar, tspan)

    def f(self, y, t):
        pass

    def G(self, y, t):
        pass


class DDEModel(Model):
    """Model defined by a system of delay differential equations
    """
    pass


class DelayItoModel(Model):
    """Model defined by a system of Ito stochastic delay differential equations
    """
    pass


class NetworkModel(Model):
    """Model consisting of many coupled instances of a sub-model connected in a
    network. The sub-models do not need to be identical.

    It is assumed that inputs from multiple sources to one target submodel
    should be summed linearly to get the overall input driving the target.
    For ODE:   dy_i/dt = f_i(y, t)  + \sum_{j=1}^{n}{coupling_ij(y, weight_j)}

    For SDE:   dy_i = f_i(y, t)dt + \sum_{j=1}^{n}{coupling_ij(y, weight_j)}
                      + G_i(y, t).dot(dW_i)

    Indexing with [i] gives access to the ith sub-model in the network.

    Attributes:
      network (array of shape (n, n)): adjacency matrix defining the network.
      timeseries: resulting timeseries: all variables of all nodes.
      output: resulting timeseries: only output variables of output nodes.
      output_nodes (optional list of int): which nodes of the network should be
        used to define the output of the overall network (default: all nodes)
    """
    def __init__(self, submodels, network, coupling_function=None,
                 independent_noise=True):
        """Construct a network model from submodels and a connectivity matrix.

        Arguments:
          submodels (sequence of Model): n subsystems comprising the nodes of
            the network. The submodels do not have to be identical. 

          network (array of shape (n, n)): Adjacency matrix defining the
            network edges as a weighted, directed graph. If network[i, j] is
            nonzero, this means subsystem i provides input to subsystem j with
            connection strength (weight) given by the value of network[i, j].

          coupling_function (callable): Function `coupling(source_y, target_y,
            weight)` How the output of a source subsystem should be coupled to
            drive the variables of a target subsystem. If `None`, the default
            is to look for a coupling function defined by models being coupled:
            `submodels[0].coupling()`

            If `None` and the submodel does not define any coupling function,
            then the fallback is to take the mean of all variables of the
            source subsystem and use that value weighted by the connection
            strength to drive all variables of the target subsystem. That is
            probably not what you want, so supply a function here. When
            defining a coupling function it should have the same signature
            as NetworkModel.coupling()

          independent_noise (optional bool): This option is only used for
            stochastic systems. If True, the noise processes driving each
            submodel will be independent (this may be suitable if the noise is
            modelling processes intrinsic to each submodel). If False, all
            submodels will share the same noise inputs (suitable in some cases
            where the noise is extrinsic).
        """
        self.submodels = [m() if (isinstance(m, type) and issubclass(m, Model))
                          else m for m in submodels] # permit classes
        self.submodels = [self._scalar_to_vector(m) for m in self.submodels]
        self._n = len(self.submodels)
        self.dimension = 0
        self._sublengths = [] # dimension of the state space of each submodel
        self._si = [0] # starting indices for each submodel, and one beyond end
        for m in self.submodels:
            self.dimension += m.dimension
            self._sublengths.append(m.dimension)
            self._si.append(self.dimension)
        self.output_nodes = range(self._n)
        super(NetworkModel, self).__init__()
        # decide whether to treat overall system as ODE, Ito or Stratonovich
        submodel_classes = []
        self.submodel_class = ODEModel # default to ODE, if no SDE submodels
        self.integrator = (integrate.odeint,)
        self.coupling_function = (self.coupling,)
        for m in self.submodels:
            if isinstance(m, NetworkModel):
                m = m.submodel_class
            if isinstance(m, ODEModel):
                submodel_classes.append(ODEModel)
            elif isinstance(m, ItoModel):
                submodel_classes.append(ItoModel)
                self.submodel_class = ItoModel
                self.integrator = (sdeint.itoint,)
            elif isinstance(m, StratonovichModel):
                submodel_classes.append(StratonovichModel)
                self.submodel_class = StratonovichModel
                self.integrator = (sdeint.stratint,)
            else:
                raise SimValueError(
                  """currently only instances of an ODEModel, ItoModel, 
                  StratonovichModel or NetworkModel are supported as 
                  subsystems of a NetworkModel""")
        if (ItoModel in submodel_classes and 
                StratonovichModel in submodel_classes):
            raise SimValueError(
              "can't use Ito and Stratonovich submodels in the same network")
        network = np.array(network)
        if network.shape != (self._n, self._n):
            raise SimValueError(
              'for %d submodels, adjacency matrix should be shape (%d,%d)' % (
              self._n, self._n, self._n))
        self.network = network
        # Validate coupling function
        test_suby0 = self.submodels[0].y0
        test_weight = 0.5
        if coupling_function is not None:
            self.coupling_function = (coupling_function,)
        elif (isinstance(self.submodels[0], Model) and
              hasattr(self.submodels[0], 'coupling') and
              callable(self.submodels[0].coupling)):
            self.coupling_function = (self.submodels[0].coupling,)
        else:
            warnings.warn('No coupling function defined. Using the default.',
                          RuntimeWarning, 1)
        try:
            test_res = self.coupling_function[0](test_suby0, test_suby0,
                                                 test_weight)
        except:
            print('Coupling function failed given arguments %s, %s, %s' % (
                  test_suby0, test_suby0, test_weight))
            raise
        if not test_res.shape == test_suby0.shape:
            raise SimValueError(
                """Coupling function must return an array with the same shape
                as the state vector y of the target system: %s""" %
                test_suby0.shape)
        self.y0 = np.concatenate([m.y0 for m in self.submodels], axis=0)
        self._independent_noise = independent_noise
        self._nsubnoises = []
        for m in self.submodels:
            self._nsubnoises.append(self._number_of_driving_noises(m))
        # determine number of driving Wiener processes for overall network
        if self._independent_noise:
            self.nnoises = sum(p for p in self._nsubnoises)
        else:
            self.nnoises = max(p for p in self._nsubnoises)

    def coupling(self, source_y, target_y, weight):
        """How to couple the output of one subsystem to the input of another.

        This is a fallback default coupling function that should usually be
        replaced with your own.

        This example coupling function takes the mean of all variables of the
        source subsystem and uses that value weighted by the connection
        strength to drive all variables of the target subsystem.

        Arguments:
          source_y (array of shape (d,)): State of the source subsystem.
          target_y (array of shape (d,)): State of target subsystem.
          weight (float): the connection strength for this connection.

        Returns:
          input (array of shape (d,)): Values to drive each variable of the
            target system.
        """
        return np.ones_like(target_y)*np.mean(source_y)*weight

    def f(self, y, t):
        """Deterministic term f of the complete network system
        dy = f(y, t)dt + G(y, t).dot(dW)

        (or for an ODE network system without noise, dy/dt = f(y, t))

        Args:
          y (array of shape (d,)): where d is the dimension of the overall
            state space of the complete network system. 

        Returns: 
          f (array of shape (d,)):  Defines the deterministic term of the
            complete network system
        """
        coupling = self.coupling_function[0]
        res = np.empty_like(self.y0)
        for j, m in enumerate(self.submodels):
            slicej = slice(self._si[j], self._si[j+1])
            target_y = y[slicej] # target node state
            res[slicej] = m.f(target_y, t) # deterministic part of submodel j
            # get indices of all source nodes that provide input to node j:
            sources = np.nonzero(self.network[:,j])[0]
            for i in sources:
                weight = self.network[i, j]
                source_y = y[slice(self._si[i], self._si[i+1])] # source state
                res[slicej] += coupling(source_y, target_y, weight)
        return res

    def G(self, y, t):
        """Noise coefficient matrix G of the complete network system
        dy = f(y, t)dt + G(y, t).dot(dW)

        (for an ODE network system without noise this function is not used)

        Args:
          y (array of shape (d,)): where d is the dimension of the overall
            state space of the complete network system. 

        Returns:
          G (array of shape (d, m)): where m is the number of independent
            Wiener processes driving the complete network system. The noise
            coefficient matrix G defines the stochastic term of the system.
        """
        if self._independent_noise:
            # then G matrix consists of submodel Gs diagonally concatenated:
            res = np.zeros((self.dimension, self.nnoises))
            offset = 0
            for j, m in enumerate(self.submodels):
                slicej = slice(self._si[j], self._si[j+1])
                ix = (slicej, slice(offset, offset + self._nsubnoises[j]))
                res[ix] = m.G(y[slicej], t) # submodel noise coefficient matrix
                offset += self._nsubnoises[j]
        else:
            # identical driving: G consists of submodel Gs stacked vertically
            res = np.empty((self.dimension, self.nnoises))
            for j, m in enumerate(self.submodels):
                slicej = slice(self._si[j], self._si[j+1])
                ix = (slicej, slice(None))
                res[ix] = m.G(y[slicej], t) # submodel noise coefficient matrix
        return res

    def integrate(self, tspan):
        if self.submodel_class is ODEModel:
            ar = self.integrator[0](self.f, self.y0, tspan)
        elif (self.submodel_class is ItoModel or
              self.submodel_class is StratonovichModel):
            ar = self.integrator[0](self.f, self.G, self.y0, tspan)
        return Timeseries(ar, tspan)

    def __len__(self):
        return len(self.submodels)

    def __getitem__(self, key):
        return self.submodels[key]

    def _number_of_driving_noises(self, submodel):
        if isinstance(submodel, ODEModel):
            return 0
        else:
            t0 = 0.0
            return submodel.G(submodel.y0, t0).shape[1]

    @property
    def output_nodes(self):
        return self._output_nodes

    @output_nodes.setter
    def output_nodes(self, seq):
        self._output_nodes = list(seq)
        self.output_vars = []
        for k in self._output_nodes:
            self.output_vars.extend((v + self._si[k]) for
                                    v in self.submodels[k].output_vars)

    def _scalar_to_vector(self, m):
        """Allow submodels with scalar equations. Convert to 1D vector systems.
        Args:
          m (Model)
        """
        if not isinstance(m.y0, numbers.Number):
            return m
        else:
            m = copy.deepcopy(m)
            t0 = 0.0
            if isinstance(m.y0, numbers.Integral):
                numtype = np.float64
            else:
                numtype = type(m.y0)
            y0_orig = m.y0
            m.y0 = np.array([m.y0], dtype=numtype)
            def make_vector_fn(fn):
                def newfn(y, t):
                    return np.array([fn(y[0], t)], dtype=numtype)
                newfn.__name__ = fn.__name__
                return newfn
            def make_matrix_fn(fn):
                def newfn(y, t):
                    return np.array([[fn(y[0], t)]], dtype=numtype)
                newfn.__name__ = fn.__name__
                return newfn
            def make_coupling_fn(fn):
                def newfn(source_y, target_y, weight):
                    return np.array([fn(source_y[0], target_y[0], weight)])
                newfn.__name__ = fn.__name__
                return newfn
            if isinstance(m.f(y0_orig, t0), numbers.Number):
                m.f = make_vector_fn(m.f)
            if hasattr(m, 'G') and isinstance(m.G(y0_orig,t0), numbers.Number):
                m.G = make_matrix_fn(m.G)
            if (hasattr(m, 'coupling') and
                    isinstance(m.coupling(y0_orig, y0_orig, 0.5),
                               numbers.Number)):
                m.coupling = make_coupling_fn(m.coupling)
            return m

    def _node_labels(self):
        return [u'node %d' % i for i in range(self._n)]

    def _reshape_timeseries(self, ts):
        """Introduce a new axis 2 that ranges across nodes of the network"""
        if np.count_nonzero(np.diff(self._sublengths)) == 0:
            # then all submodels have the same dimension, so can reshape array
            # in place without copying data:
            subdim = self.dimension / self._n
            shp = list(ts.shape)
            shp[1] = self._n
            shp.insert(2, subdim)
            ts = ts.reshape(tuple(shp)).swapaxes(1, 2)
            # label variables only if all sub-models agree on the labels:
            all_var_labels = [m.labels for m in self.submodels]
            var_labels = all_var_labels[0]
            if all(v == var_labels for v in all_var_labels[1:]):
                ts.labels[1] = var_labels
            ts.labels[2] = self._node_labels()
            return ts
        else:
            # will pad with zeros for submodels with less variables
            subdim = max(self._sublengths)
            shp = list(ts.shape)
            shp[1] = subdim
            shp.insert(2, self._n)
            ar = np.zeros(shp)
            labels = ts.labels
            labels.insert(2, self._node_labels())
            for k in range(self._n):
                sl = slice(self._si[k], self._si[k+1])
                ar[:,:,k,...] = ts[:,sl,...]
            return Timeseries(ar, ts.tspan, labels)

    def _reshape_output(self, ts):
        """Introduce a new axis 2 that ranges across nodes of the network"""
        subodim = len(self.submodels[0].output_vars)
        shp = list(ts.shape)
        shp[1] = subodim
        shp.insert(2, self._n)
        ts = ts.reshape(tuple(shp))
        ts.labels[2] = self._node_labels()
        return ts


class Simulation(object):
    """Represents a single simulation of a system, the parameter settings that
    were used and the resulting time series.

    Attributes:
      system (Model): The dynamical system that was simulated, with parameters.
      tspan (array): The sequence of time points that was simulated
      timeseries (array of shape (len(tspan), len(y0))): 
        Multivariate time series of full simulation results.
      output: Some function of the simulated timeseries, for example a 
        univariate time series of a single output variable. 
    """
    def __init__(self, system, T=60.0, dt=0.005, integrator=None):
        """
        Args:
          system (Model): The dynamical system to simulate. (Here you can
            provide either a Model class or a Model instance. If a class is
            provided then the class will be instantiated to create this
            simulation, and any random variables in the model's parameters
            will be instantiated with specific values from the distribution.)
          T (Number, optional): Total length of time to simulate, in seconds.
          dt (Number, optional): Timestep for numerical integration.
          integrator (callable, optional): Which numerical integration
            algorithm to use. If None, the model's default algorithm will be
            used. The integrator function should accept the same arguments as
            the sdeint library, e.g. y = integrator(f, y0, tspan) for an ODE or
            y = integrator(f, G, y0, tspan) for a SDE.
        """
        if isinstance(system, type):
            self.system = system()
        else:
            self.system = system
        if integrator is not None:
            self.system.integrator = [integrator]
        self.T = T
        self.dt = dt
        self._timeseries = None

    def compute(self):
         tspan = np.arange(0, self.T + self.dt, self.dt)
         if hasattr(self.system, 'labels'):
             labels = [None, self.system.labels]
         else:
             labels = [None, None]
         ar = self.system.integrate(tspan)
         self._timeseries = Timeseries(ar, tspan, labels)

    @property
    def timeseries(self):
        """Simulated time series"""
        if self._timeseries is None:
            self.compute()
        if isinstance(self.system, NetworkModel):
            return self.system._reshape_timeseries(self._timeseries)
        else:
            return self._timeseries

    @property
    def output(self):
        """Simulated model output"""
        if self._timeseries is None:
            self.compute()
        output = self._timeseries[:, self.system.output_vars]
        if isinstance(self.system, NetworkModel):
            return self.system._reshape_output(output)
        else:
            return output


class MultipleSim(object):
    """Represents multiple simulations, on a single host

    Like a list, indexing with [i] gives access to the ith simulation

    Attributes:
      timeseries: resulting timeseries: all variables of all simulations
      output: resulting timeseries: output variables of all simulations
    """
    def __init__(self, systems, T=60.0, dt=0.005, integrator=None):
        """
        Args:
          systems: sequence of Model instances that should be simulated.
          T: total length of time to simulate, in seconds.
          dt: timestep for numerical integration.
          integrator (callable, optional): Which numerical integration
            algorithm to use. If None, the model's default algorithm will be
            used. The integrator function should accept the same arguments as
            the sdeint library, e.g. y = integrator(f, y0, tspan) for an ODE or
            y = integrator(f, G, y0, tspan) for a SDE.
        """
        self.T = T
        self.dt = dt
        self.sims = [Simulation(s, T, dt, integrator) for s in systems]

    def compute(self):
        for s in self.sims:
            s.compute()

    def __len__(self):
        return len(self.sims)

    def __getitem__(self, key):
        return self.sims[key]

    def _ids(self):
        return tuple(id(sim) for sim in self.sims)
 
    def __repr__(self):
        s = '%s([\n' % self.__class__.__name__
        for i in range(len(self)):
            if i:
                s += ',\n'
            s += '<%s at %s>' % (self.sims[i].__class__.__name__,
                                 hex(self._ids()[i]))
        return s + '])'

    def _node_labels(self):
        return ['node %d' % i for i in range(len(self.sims))]
   
    @property
    def timeseries(self):
        """Rank 3 array representing multiple time series. Axis 0 is time,
        axis 1 ranges across all dynamical variables in a single simulation,
        axis 2 ranges across different simulation instances."""
        subts = [s.timeseries for s in self.sims]
        sub_ndim = subts[0].ndim
        if sub_ndim is 1:
            subts = [distob.expand_dims(ts, 1) for ts in subts]
            sub_ndim += 1
        nodeaxis = sub_ndim
        subts = [distob.expand_dims(ts, nodeaxis) for ts in subts]
        ts = subts[0].concatenate(subts[1:], axis=nodeaxis)
        ts.labels[nodeaxis] = self._node_labels()
        return ts

    @property
    def output(self):
        """Rank 3 array representing output time series. Axis 0 is time, 
        axis 1 ranges across output variables of a single simulation, 
        axis 2 ranges across different simulation instances."""
        subts = [s.output for s in self.sims]
        sub_ndim = subts[0].ndim
        if sub_ndim is 1:
            subts = [distob.expand_dims(ts, 1) for ts in subts]
            sub_ndim += 1
        nodeaxis = sub_ndim
        subts = [distob.expand_dims(ts, nodeaxis) for ts in subts]
        ts = subts[0].concatenate(subts[1:], axis=nodeaxis)
        ts.labels[nodeaxis] = self._node_labels()
        return ts


@distob.proxy_methods(Simulation)
class RemoteSimulation(distob.Remote, Simulation):
    """Local object representing a remote Simulation"""
    def __init__(self, ref):
        """Make a RemoteSimulation to access an already-existing Simulation 
        object, which may be on a remote engine.

        Args:
          ref (Ref): reference to a Simulation to be controlled by this proxy
        """
        super(RemoteSimulation, self).__init__(ref)

    def compute(self):
        """Start the computation process asynchronously"""
        from distob import methodcall
        methodcall(self, 'compute', prefer_local=False, block=False)

    def __repr__(self):
        return '<%s at %s on engine %d>' % (
                self.__class__.__name__, hex(self._id[0]), self._id[1])


@distob.proxy_methods(MultipleSim, include_underscore=('__getitem__', '_ids'))
class RemoteMultipleSim(distob.Remote, MultipleSim):
    """Local object representing a remote MultipleSim"""
    def __init__(self, ref):
        """Make a RemoteMultipleSim to access an already-existing MultipleSim
        object, which may be on a remote engine.

        Args:
          ref (Ref): reference to a MultipleSim to be controlled by this proxy
        """
        super(RemoteMultipleSim, self).__init__(ref)

    def compute(self):
        """Start the computation process asynchronously"""
        from distob import methodcall
        methodcall(self, 'compute', prefer_local=False, block=False)

    def __repr__(self):
        s = '%s([\n' % self.__class__.__name__
        for i in range(len(self)):
            if i:
                s += ',\n'
            s += '<RemoteSimulation at %s on engine %d>' % (
                    hex(self._ids()[i]), self._id[1])
        return s + '])'


class DistSim(object):
    """Represents multiple simulations distributed on multiple compute engines

    Like a list, indexing with [i] gives access to the ith simulation

    Attributes:
      timeseries: resulting timeseries: all variables of all simulations
      output: resulting timeseries: output variables of all simulations
    """
    def __init__(self, systems, T=60.0, dt=0.005, integrator=None):
        """
        Args:
          systems: sequence of Model instances that should be simulated.
          T: total length of time to simulate, in seconds.
          dt: timestep for numerical integration.
          integrator (callable, optional): Which numerical integration
            algorithm to use. If None, the model's default algorithm will be
            used. The integrator function should accept the same arguments as
            the sdeint library, e.g. y = integrator(f, y0, tspan) for an ODE or
            y = integrator(f, G, y0, tspan) for a SDE.
        """
        self.T = T
        self.dt = dt
        self._n = len(systems)
        n = self._n
        if n == 1:
            return distob.scatter(Simulation(systems[0], T, dt, integrator))
        if distob.engine is None:
            distob.setup_engines()
        ne = distob.engine.nengines
        blocksize = ((n - 1) // ne) + 1
        if blocksize > n:
            blocksize = n
        self._subsims = []
        sublengths = []
        si = [0]
        low = 0
        for i in range(0, n // blocksize):
            high = low + blocksize
            self._subsims.append(
                    MultipleSim(systems[low:high], T, dt, integrator))
            sublengths.append(blocksize)
            si.append(si[-1] + sublengths[-1])
            low += blocksize
        if n % blocksize != 0:
            high = low + (n % blocksize)
            self._subsims.append(
                    MultipleSim(systems[low:high], T, dt, integrator))
            sublengths.append(high - low)
            si.append(si[-1] + sublengths[-1])
        self._sublengths = tuple(sublengths)
        self._si = tuple(si)
        # a surragate ndarray to help with slicing
        self._placeholders = np.arange(n, dtype=int)
        # distribute them on cluster and start computation:
        self._subsims = distob.scatter(self._subsims)
        for rms in self._subsims:
            rms.compute()

    def __len__(self):
        return self._n

    def _tosub(self, ix):
        """Given an integer index ix into the list of sims, returns the pair
        (s, m) where s is the relevant subsim and m is the subindex into s.
        So self[ix] == self._subsims[s][m]
        """
        N = self._n
        if ix >= N or ix < -N:
            raise IndexError(
                    'index %d out of bounds for list of %d sims' % (ix, N))
        if ix < 0:
            ix += N
        for s in range(0, self._n):
            if self._si[s + 1] - 1 >= ix:
                break
        m = ix - self._si[s]
        return s, m

    def _tosubs(self, ixlist):
        """Maps a list of integer indices to sub-indices.
        ixlist can contain repeated indices and does not need to be sorted.
        Returns pair (ss, ms) where ss is a list of subsim numbers and ms is a
        list of lists of subindices m (one list for each subsim in ss).
        """
        n = len(ixlist)
        N = self._n
        ss = []
        ms = []
        if n == 0:
            return ss, ms
        j = 0 # the position in ixlist currently being processed
        ix = ixlist[j]
        if ix >= N or ix < -N:
            raise IndexError(
                    'index %d out of bounds for list of %d sims' % (ix, N))
        if ix < 0:
            ix += N
        while j < n:
            for s in range(0, self._n):
                low = self._si[s]
                high = self._si[s + 1]
                if ix >= low and ix < high:
                    ss.append(s)
                    msj = [ix - low]
                    j += 1
                    while j < n:
                        ix = ixlist[j]
                        if ix >= N or ix < -N:
                            raise IndexError(
                              'index %d out of bounds for list of %d sims' % (
                                ix, N))
                        if ix < 0:
                            ix += N
                        if ix < low or ix >= high:
                            break
                        msj.append(ix - low)
                        j += 1
                    ms.append(msj)
                if ix < low:
                    break
        return ss, ms

    def __getitem__(self, key):
        ixlist = self._placeholders[key]
        if isinstance(ixlist, numbers.Number):
            s, i = self._tosub(ixlist)
            return self._subsims[s][i]
        remotesims = []
        ss, ms = self._tosubs(ixlist)
        for s, m in zip(ss, mm):
            remotesims.append(self._subsims[s][m])
        if len(remotesims) is 1:
            return remotesims[0]
        else:
            return remotesims

    def __repr__(self):
        s = '%s([\n' % self.__class__.__name__
        for i in range(len(self._subsims)):
            if i:
                s += ',\n'
            s += '<%s at %s on engine %d (sims %d to %d)>' % (
                    self._subsims[i].__class__.__name__,
                    hex(self._subsims[i]._id[0]), self._subsims[i]._id[1],
                    self._si[i], self._si[i+1] - 1)
        return s + '])'

    def _node_labels(self):
        return ['node %d' % i for i in range(self._n)]

    @property
    def timeseries(self):
        """Rank 3 array representing multiple time series. Axis 0 is time, 
        axis 1 ranges across all dynamical variables in a single simulation, 
        axis 2 ranges across different simulation instances."""
        subts = [rms.timeseries for rms in self._subsims]
        distaxis = subts[0].ndim - 1
        return DistTimeseries(subts, distaxis, self._node_labels())

    @property
    def output(self):
        """Rank 3 array representing output time series. Axis 0 is time, 
        axis 1 ranges across output variables of a single simulation, axis 2 
        ranges across different simulation instances."""
        subts = [rms.output for rms in self._subsims]
        distaxis = subts[0].ndim - 1
        return DistTimeseries(subts, distaxis, self._node_labels())


class RepeatedSim(DistSim):
    """Independent simulations of the same model multiple times, with results.

    Like a list, indexing the object with [i] gives access to the ith simulation

    Attributes:
      modelclass: the Model class common to all the simulations
      timeseries: resulting timeseries: all variables of all simulations
      output: resulting timeseries: output variables of all simulations
    """
    def __init__(self, system, T=60.0, dt=0.005, repeat=1, identical=True,
                 integrator=None):
        """
        Args:
          system (Model): The dynamical system to simulate. (Here you can
            provide either a Model class or a Model instance. If a class is
            provided then the class will be instantiated repeatedly to create
            the simulations, and any random variables in the model's parameters
            will be instantiated with specific values.

          T (optional): total length of time to simulate, in seconds.

          dt (optional): timestep for numerical integration.

          repeat (int, optional): number of repeated simulations of the model.

          identical (bool, optional): Whether repeated simulations of the model
            should use the same parameters. If identical=False, for each
            simulation different values will be drawn randomly from
            distributions specified in the model (assuming the model has some
            parameters specified as distributions). If identical=True values
            will be chosen once then the same values used in each simulation.

          integrator (callable, optional): Which numerical integration
            algorithm to use. If None, the model's default algorithm will be
            used. The integrator function should accept the same arguments as
            the sdeint library, e.g. y = integrator(f, y0, tspan) for an ODE or
            y = integrator(f, G, y0, tspan) for a SDE.
        """
        if isinstance(system, type):
            self.modelclass = system # class
            model = self.modelclass() # instance
        else:
            self.modelclass = type(system) # class
            model = system # instance
        if identical is True:
            systems = [copy.deepcopy(model) for i in range(repeat)]
        else:
            if isinstance(model, NetworkModel):
                network = model.network
                coupling = model.coupling_function[0]
                independent_noise = model._independent_noise
                systems = []
                for i in range(repeat):
                    # generate new, non-identical instances of each sub-model
                    submodels = [type(m)() for m in model.submodels]
                    networkmodel = self.modelclass(submodels, network,
                                                   coupling, independent_noise)
                    systems.append(networkmodel)
            else:
                systems = [self.modelclass() for i in range(repeat)]
        super(RepeatedSim, self).__init__(systems, T, dt, integrator)

    def _node_labels(self):
        return ['repetition %d' % i for i in range(self._n)]


class ParameterSim(DistSim):
    """Separate simulations of a model to explore a lattice of different
    parameter values."""
    pass


def newsim(f, G, y0, name='NewModel', modelType=ItoModel, T=60.0, dt=0.005, repeat=1, identical=True):
    """Make a simulation of the system defined by functions f and G.

    dy = f(y,t)dt + G(y,t).dW with initial condition y0
    This helper function is for convenience, making it easy to define 
    one-off simulations interactively in ipython.

    Args:
      f: callable(y, t) (defined in global scope) returning (n,) array
        Vector-valued function to define the deterministic part of the system 
      G: callable(y, t) (defined in global scope) returning (n,m) array
        Optional matrix-valued function to define noise coefficients of an Ito
        SDE system.
      y0 (array):  Initial condition 
      name (str): Optional class name for the new model
      modelType (type): The type of model to simulate. Must be a subclass of
        nsim.Model, for example nsim.ODEModel, nsim.ItoModel or 
        nsim.StratonovichModel. The default is nsim.ItoModel.
      T: Total length of time to simulate, in seconds.
      dt: Timestep for numerical integration.
      repeat (int, optional)
      identical (bool, optional)

    Returns: 
      Simulation

    Raises:
      SimValueError, SimTypeError
    """
    NewModel = newmodel(f, G, y0, name, modelType)
    if repeat == 1:
        return Simulation(NewModel(), T, dt)
    else:
        return RepeatedSim(NewModel, T, dt, repeat, identical)


def newmodel(f, G, y0, name='NewModel', modelType=ItoModel):
    """Use the functions f and G to define a new Model class for simulations. 

    It will take functions f and G from global scope and make a new Model class
    out of them. It will automatically gather any globals used in the definition
    of f and G and turn them into attributes of the new Model.

    Args:
      f: callable(y, t) (defined in global scope) returning (n,) array
         Scalar or vector-valued function to define the deterministic part
      G: callable(y, t) (defined in global scope) returning (n,m) array
         Optional scalar or matrix-valued function to define noise coefficients
         of a stochastic system. This should be ``None`` for an ODE system.
      y0 (Number or array): Initial condition
      name (str): Optional class name for the new model
      modelType (type): The type of model to simulate. Must be a subclass of
        nsim.Model, for example nsim.ODEModel, nsim.ItoModel or 
        nsim.StratonovichModel. The default is nsim.ItoModel.

    Returns: 
      new class (subclass of Model)

    Raises:
      SimValueError, SimTypeError
    """
    if not issubclass(modelType, Model):
        raise SimTypeError('modelType must be a subclass of nsim.Model')
    if not callable(f) or (G is not None and not callable(G)):
        raise SimTypeError('f and G must be functions of y and t.')
    if G is not None and f.__globals__ is not G.__globals__:
        raise SimValueError('f and G must be defined in the same place')
    # TODO: validate that f and G are defined at global scope.
    # TODO: Handle nonlocals used in f,G so that we can lift this restriction.
    if modelType is ODEModel and G is not None and not np.all(G == 0.0):
        raise SimValueError('For an ODEModel, noise matrix G should be None')
    if G is None or modelType is ODEModel:
        newclass = type(name, (ODEModel,), dict())
        setattr(newclass, 'f', staticmethod(__clone_function(f, 'f')))
    else:
        newclass = type(name, (modelType,), dict())
        setattr(newclass, 'f', staticmethod(__clone_function(f, 'f')))
        setattr(newclass, 'G', staticmethod(__clone_function(G, 'G')))
    setattr(newclass, 'y0', copy.deepcopy(y0))
    # For any global that is used by the functions f or G, create a 
    # corresponding attribute in our new class.
    globals_used = [x for x in f.__globals__ if (x in f.__code__.co_names or 
        G is not None and x in G.__code__.co_names)]
    for x in globals_used:
        if G is None:
            setattr(newclass, x, __AccessDict(x, newclass.f.__globals__))
        else:
            setattr(newclass, x, __AccessDicts(x, newclass.f.__globals__, 
                                                  newclass.G.__globals__))
    # Put the new class into namespace __main__ (to cause dill to pickle it)
    newclass.__module__ = '__main__'
    import __main__
    __main__.__dict__[name] = newclass 
    return newclass


class __AccessDict(object):
    """A descriptor class representing a value held in a dict.
      k (object): The common key to access. 
      d (dict)
    """
    def __init__(self, k, d):
        self.k = k
        self.d = d
    def __get__(self, obj, objtype):
        return self.d[self.k]
    def __set__(self, obj, val):
        self.d[self.k] = val


class __AccessDicts(object):
    """A descriptor class representing a common value held in two dicts.
      k (object): The common key to access. 
      d1, d2 (dict):  (Invariant: d1[k] is d2[k])
    """
    def __init__(self, k, d1, d2):
        self.k = k
        self.d1 = d1
        self.d2 = d2
    def __get__(self, obj, objtype):
        return self.d1[self.k]
    def __set__(self, obj, val):
        self.d1[self.k] = val
        self.d2[self.k] = val


def __clone_function(f, name=None):
    """Make a new version of a function that has its own independent copy 
    of any globals that it uses directly, and has its own name. 
    All other attributes are assigned from the original function.

    Args:
      f: the function to clone
      name (str):  the name for the new function (if None, keep the same name)

    Returns:
      A copy of the function f, having its own copy of any globals used

    Raises:
      SimValueError
    """
    if not isinstance(f, types.FunctionType):
        raise SimTypeError('Given parameter is not a function.')
    if name is None:
        name = f.__name__
    newglobals = f.__globals__.copy()
    globals_used = [x for x in f.__globals__ if x in f.__code__.co_names]
    for x in globals_used:
        gv = f.__globals__[x]
        if isinstance(gv, types.FunctionType):
            # Recursively clone any global functions used by this function.
            newglobals[x] = __clone_function(gv)
        elif isinstance(gv, types.ModuleType):
            newglobals[x] = gv
        else:
            # If it is something else, deep copy it.
            newglobals[x] = copy.deepcopy(gv)
    newfunc = types.FunctionType(
        f.__code__, newglobals, name, f.__defaults__, f.__closure__)
    return newfunc



"""This shows the simpler syntax for interactive use.  

simulating Ito SDE   dy = lam y dt + sigma (0.2 + y) dW
"""

import nsim

lam = -1.0
sigma = 0.3
y0 = 0.0

def f(y, t):
    return lam * y

def G(y, t):
    return sigma * (0.2 + y)


Model = nsim.newmodel(f, G, y0)

s = nsim.Simulation(Model, T=400.0)
s.output.plot()



"""ODE example:    d^2y/dt^2 = (0.5 sin(0.3 t) - 0.1) dy/dt - 3 y

writing as a two-dimensional system:

                   dy1/dt = y2
                   dy2/dt = (0.5 sin(0.3 t) - 0.1) y2 - 3 y1
"""
import nsim
import numpy as np



class ExampleODE(nsim.ODEModel):

    y0 = np.array([ 1.0, 
                    1.0 ])

    def f(self, y, t):
        return np.array([ y[1], 
                          (0.5*np.sin(0.3*t) - 0.1)*y[1] - 3*y[0] ])
                            


s = nsim.Simulation(ExampleODE, T=120.0)
s.timeseries.plot()



"""An empty file where tests should be
"""

import pytest
import nsim
import numpy as np


def test_slicing():
    ar = np.random.randn(3, 4, 5)
    labels = [None, ['c0', 'c1', 'c2', 'c3'], ['n0', 'n1', 'n2', 'n3', 'n4']]
    ts = nsim.Timeseries(ar, labels=labels)
    # slicing can add a new axis
    assert(ts[:, np.newaxis, :, :].shape == (3, 1, 4, 5))
    assert(isinstance(ts[:, np.newaxis, :, :], nsim.Timeseries))
    assert(ts[np.newaxis, :, :, :].shape == (1, 3, 4, 5))
    assert(not isinstance(ts[np.newaxis, :, :, :], nsim.Timeseries))


def test_slicing_labels():
    ar = np.random.randn(3, 4, 5)
    labels = [None, ['c0', 'c1', 'c2', 'c3'], ['n0', 'n1', 'n2', 'n3', 'n4']]
    ts = nsim.Timeseries(ar, labels=labels)
    assert(ts[:, :, :].labels == ts.labels)
    assert(ts[:, 2, :].labels[1][4] == 'n4')
    assert(len(ts[:, :, 3].labels) is 2)
    assert(ts[:, :, 3].labels[1][3] == 'c3')
    assert(ts[:, 1, 1].labels == [None])
    assert(all(np.equal(ts[:, 1, 1].tspan, ts.tspan)))
    assert(ts[..., np.newaxis].labels[3] is None)
    assert(ts[..., np.newaxis, :].labels[3][0] == 'n0')
    partial_labels = [None, None, ['n0', 'n1', 'n2', 'n3', 'n4']]
    ts1 = nsim.Timeseries(ar, labels=partial_labels)
    assert(ts1[:, 2, :].labels[1][4] == 'n4')
    assert(len(ts1[:, :, 3].labels) is 2)
    assert(ts1[:, :, 3].labels[1] is None)


def test_reshape():
    ar = np.random.randn(4, 4, 4)
    labels = [None, ['c0', 'c1', 'c2', 'c3'], ['n0', 'n1', 'n2', 'n3']]
    ts = nsim.Timeseries(ar, labels=labels)
    assert(not isinstance(ts.reshape((2, 2, 4, 4)), nsim.Timeseries))
    ts1 = ts.reshape((-1, 4, 2, 2))
    assert(all(np.equal(ts1.tspan, ts.tspan)))
    assert(ts1.labels[1] == ts.labels[1])
    assert(ts1.labels[2] is None)



import numpy as np
from scipy.signal import wavelets, convolve


def cwtmorlet(points, width):
    """complex morlet wavelet function compatible with scipy.signal.cwt
    Parameters: points: int
                    Number of points in `vector`.
                width: scalar
                    Width parameter of wavelet.
                    Equals (sample rate / fundamental frequency of wavelet)
    Returns: `vector`: complex-valued ndarray of shape (points,)
    """
    omega = 5.0
    s = points / (2.0 * omega * width)
    return wavelets.morlet(points, omega, s, complete=True)


# modified cwt for less accuracy and more speed (use 3*width, not 10*width)
def roughcwt(data, wavelet, widths):
    """
    Continuous wavelet transform.

    Performs a continuous wavelet transform on `data`,
    using the `wavelet` function. A CWT performs a convolution
    with `data` using the `wavelet` function, which is characterized
    by a width parameter and length parameter.

    Parameters
    ----------
    data : (N,) ndarray
        data on which to perform the transform.
    wavelet : function
        Wavelet function, which should take 2 arguments.
        The first argument is the number of points that the returned vector
        will have (len(wavelet(width,length)) == length).
        The second is a width parameter, defining the size of the wavelet
        (e.g. standard deviation of a gaussian). See `ricker`, which
        satisfies these requirements.
    widths : (M,) sequence
        Widths to use for transform.

    Returns
    -------
    cwt: (M, N) ndarray
        Will have shape of (len(data), len(widths)).

    Notes
    -----
    >>> length = min(10 * width[ii], len(data))
    >>> cwt[ii,:] = scipy.signal.convolve(data, wavelet(length,
    ...                                       width[ii]), mode='same')

    Examples
    --------
    >>> from scipy import signal
    >>> sig = np.random.rand(20) - 0.5
    >>> wavelet = signal.ricker
    >>> widths = np.arange(1, 11)
    >>> cwtmatr = signal.cwt(sig, wavelet, widths)

    """
    out_dtype = wavelet(widths[0], widths[0]).dtype
    output = np.zeros([len(widths), len(data)], dtype=out_dtype)
    for ind, width in enumerate(widths):
        wavelet_data = wavelet(min(3 * width, len(data)), width)
        output[ind, :] = convolve(data, wavelet_data,
                                              mode='same')
    return output



# Copyright 2015 Matthew J. Aburn
# 
# This program is free software: you can redistribute it and/or modify 
# it under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License, or 
# (at your option) any later version. See <http://www.gnu.org/licenses/>.

"""
Miscelaneous analyses that apply to an ensemble of many time series
"""
import distob
from nsim import analyses1
import numpy as np


def first_return_times(dts, c=None, d=0.0):
    """For an ensemble of time series, return the set of all time intervals
    between successive returns to value c for all instances in the ensemble.
    If c is not given, the default is the mean across all times and across all
    time series in the ensemble.

    Args:
      dts (DistTimeseries)

      c (float): Optional target value (default is the ensemble mean value)

      d (float): Optional min distance from c to be attained between returns

    Returns:
      array of time intervals (Can take the mean of these to estimate the
      expected first return time for the whole ensemble)
    """
    if c is None:
        c = dts.mean()
    vmrt = distob.vectorize(analyses1.first_return_times)
    all_intervals = vmrt(dts, c, d)
    if hasattr(type(all_intervals), '__array_interface__'):
        return np.ravel(all_intervals)
    else:
        return np.hstack([distob.gather(ilist) for ilist in all_intervals])



# Copyright 2014 Matthew J. Aburn
# 
# This program is free software: you can redistribute it and/or modify 
# it under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License, or 
# (at your option) any later version. See <http://www.gnu.org/licenses/>.

"""
classes:

  Timeseries   numpy array with extra methods for time series analyses
"""

from __future__ import absolute_import
import numpy as np
from collections import Sequence
import warnings
import copy
import types
import numbers

# types for compatibility across python 2 and 3
_SliceType = type(slice(None))
_EllipsisType = type(Ellipsis)
_TupleType = type(())
_NewaxisType = type(np.newaxis)

warnings.filterwarnings('ignore', 'comparison to `None`', FutureWarning)


class Timeseries(np.ndarray):
    """A numpy array with extra methods for applying time series analyses.
    It is an array of up to 3 dimensions.

    axis 0 ranges across values of the system at different points in time.
           For a single variable time series this is the only axis needed.

    axis 1, if present, ranges across the different variables/channels of a 
            multivariate time series, at a single node.

    axis 2, if present, ranges across different nodes of a network simulation 
            or repeated simulation.                  

    Thus the shape of a Timeseries array is (N, n, m) where N is the 
    total number of time points, n is the number of variables or channels of a
    single node and m is the number of nodes.

    Slice by array index:  timeseries[103280:104800]
    Slice by time in seconds:  timeseries.t[10.0:20.5]

    Methods:
      All functions defined in the package nsim.analyses1 are available as 
      methods of the Timeseries  (as well as the usual methods of numpy arrays)

      Your own analysis functions can also be included by calling
      Timeseries.add_analyses('file.py')
    
    Attributes:
      tspan: An array of shape (N,) giving the time in seconds at each 
        time point along axis 0 of the array. tspan will always remain sorted.
        The time points are not necessarily evenly spaced.

      t: Allows slicing by time: timeseries.t[starttime:endtime, ...]

      labels (list): For each axis>0, optionally gives names to label 
        points along the axis. (labels[0] is always None as axis 0 is the
        time axis, labeled by numbers in `tspan` rather than by strings) 
        For i>0, labels[i] is a list of str of length shape(timeseries)[i]
        giving names to each position along axis i. For example labels[1] can 
        hold names for each variable or channel in a multivariate timeseries.
    """
    def __new__(cls, input_array, tspan=None, labels=None, fs=None):
        """Create a new Timeseries from an ordinary numpy array.
        Args:
          input_array (ndarray): the timeseries data.
          tspan (ndarray, optional): An array of shape (N,) giving the time 
            in seconds at each time point along axis 0 of the array. tspan will
            always remain sorted. Time points are not necessarily evenly spaced
          labels (list, optional): For each axis>0, optionally give names 
            to label points along the axis. (labels[0] must be None).
          fs (scalar, optional): sample rate. Can be given instead of tspan if
            all time points are evenly spaced.
        """
        #print('In __new__')
        obj = np.asarray(input_array).view(cls)
        if tspan is not None:
            obj.tspan = tspan
        elif isinstance(input_array, Timeseries):
            obj.tspan = input_array.tspan
        else:
            if fs is None:
                fs = 1.0
            n = len(input_array)
            obj.tspan = np.linspace(0.0, 1.0*(n-1)/fs, n, endpoint=True)
        if isinstance(input_array, Timeseries) and labels is None:
            obj.labels = input_array.labels
        else:
            obj.labels = labels
        if obj.labels is None:
            obj.labels = [None] * obj.ndim
        if len(obj.labels) < obj.ndim:
            obj.labels.extend([None] * (obj.ndim - len(obj.labels)))
        assert(obj.labels[0] is None) # time axis should not have string labels
        for i, seq in enumerate(obj.labels):
            if seq is not None:
                if len(seq) != obj.shape[i]:
                    raise ValueError(
                        '%d labels given for axis %d of length %d' % (
                        len(seq), i, obj.shape[i]))
                obj.labels[i] = list(seq)
        obj.t = _Timeslice(obj)
        return obj

    def __init__(self, input_array, tspan=None, labels=None, fs=None):
        #print('In __init__')
        pass

    @classmethod
    def add_analyses(cls, source):
        """Dynamically add new analysis methods to the Timeseries class.
        Args:
          source: Can be a function, module or the filename of a python file.
            If a filename or a module is given, then all functions defined 
            inside not starting with _ will be added as methods.

        The only restriction on the functions is that they can accept a 
        Timeseries as their first argument. So existing functions that 
        take a ndarray or array or even a list will usually also work.
        """
        if isinstance(source, types.FunctionType):
            _add_single_method(source.__name__, source)
        else:
            if isinstance(source, types.ModuleType):
                mod = source
            elif isinstance(source, types.StringTypes):
                import os
                import imp
                path = os.path.abspath(source)
                if os.path.isfile(path) and path[-3:] == '.py':
                    dir, file = os.path.split(path)
                    name = file[:-3]
                    module_info = imp.find_module(name, [dir])
                    mod = imp.load_module('nsim.' + name, *module_info)
                elif (os.path.isdir(path) and 
                        '__init__.py' in os.listdir(path)):
                    module_info = imp.find_module('__init__', [path])
                    name = os.path.basename(path)
                    mod = imp.load_module('nsim.' + name, *module_info)
                else:
                    raise Error('"%s" is not a file or directory' % source)
            else:
                raise ValueError('`source` argument not a function or module')
            for name, obj in mod.__dict__.items():
                if name[0] != '_' and isinstance(obj, types.FunctionType):
                    cls._add_single_method(name, obj)

    @classmethod
    def _add_single_method(cls, name, f):
        import sys
        setattr(cls, name, f)
        #If using RemoteTimeseries, add method to that class too:
        if 'distob' in sys.modules:
            import distob
            if distob.engine is not None and cls in distob.engine.proxy_types:
                RemoteClass = distob.engine.proxy_types[cls]
                remotemethod = distob._make_proxy_method(name, f.__doc__)
                setattr(RemoteClass, name, remotemethod)
                # also update class definitions on remote engines
                if isinstance(distob.engine, distob.ObjectHub):
                    dv = distob.engine._dv
                    dv.targets = 'all'
                    def remote_update(name, Class, f, remotemethod):
                        setattr(Class, name, f)
                        RemoteClass = distob.engine.proxy_types[Class]
                        setattr(RemoteClass, name, remotemethod)
                    ars = dv.apply(remote_update, real_type, proxy_type)
                    dv.wait(ars)
                    for ar in ars:
                        if not ar.successful():
                            raise ar.r

    def __array_finalize__(self, obj):
        #print('In __array_finalize__, obj is type ' + str(type(obj)))
        if obj is None:
            return
        if isinstance(obj, self.__class__):
            if obj.shape is () or obj.shape is not () and len(self) ==len(obj):
                self.tspan = obj.tspan
                self.labels = obj.labels
                self.t = _Timeslice(self)

    def __array_prepare__(self, in_arr, context=None):
        #print('In __array_prepare__')
        return super(Timeseries, self).__array_prepare__(in_arr, context)

    def __array_wrap__(self, out_arr, context=None):
        #print('In __array_wrap__')
        return super(Timeseries, self).__array_wrap__(out_arr, context)

    def __getitem__(self, index):
        """When a Timeseries is sliced, tspan will be sliced in the
        same way as axis 0 of the Timeseries, and labels[i] will be sliced in
        the same way as axis i of the Timeseries.
        If the resulting array is not a Timeseries (for example if the
        time axis is sliced to a scalar value) then returns an ndarray.
        """
        new_array = np.asarray(self).__getitem__(index)
        cur_labels = self.labels
        cur_shape = self.shape # current values, may be updated by np.newaxis
        cur_ndim = self.ndim
        if isinstance(index, np.ndarray) and index.dtype.type is np.bool_:
            raise Error('indexing by boolean array not yet implemented')
        if not isinstance(index, Sequence):
            index = (index,) + (slice(None),)*(self.ndim - 1)
        ix_types = tuple(type(x) for x in index)
        if (np.ndarray in ix_types or
                (not isinstance(index, _TupleType) and
                    _NewaxisType not in ix_types and
                    _EllipsisType not in ix_types and
                    _SliceType not in ix_types) or
                any(issubclass(T, Sequence) for T in ix_types)):
            basic_slicing = False
        else:
            basic_slicing = True
        # Apply any ellipses
        while _EllipsisType in ix_types:
            pos = ix_types.index(_EllipsisType)
            m = (self.ndim + ix_types.count(_NewaxisType) - len(index) + 1)
            index = index[:pos] + (slice(None),)*m + index[(pos+1):]
            ix_types = tuple(type(x) for x in index)
        # Apply any np.newaxis
        while _NewaxisType in ix_types:
            pos = ix_types.index(type(np.newaxis))
            if pos is 0:
                # prepended an axis: no longer a Timeseries
                return np.asarray(new_array)
            index = index[:pos] + (slice(None),) + index[(pos+1):]
            cur_labels = cur_labels[:pos] + [None] + cur_labels[pos:]
            cur_shape = cur_shape[:pos] + (1,) + cur_shape[pos:]
            cur_ndim = len(cur_shape)
            ix_types = tuple(type(x) for x in index)
        index = tuple(index) + (slice(None),)*(cur_ndim - len(index))
        if len(index) > cur_ndim:
            raise IndexError('too many indices for array')
        if basic_slicing:
            new_tspan = self.tspan[index[0]]
            if (not isinstance(new_tspan, np.ndarray) or 
                    new_tspan.shape is () or
                    new_tspan.shape[0] is 0):
                # axis 0 has been sliced away: result is no longer a Timeseries
                return np.asarray(new_array)
            new_labels = [None]
            for i in range(1, cur_ndim):
                # make temp ndarray of labels to ensure correct slicing
                if cur_labels[i] is None:
                    labelarray = np.array([''] * cur_shape[i])
                else:
                    labelarray = np.array(cur_labels[i])
                newlabelarray = labelarray[index[i]]
                if newlabelarray.shape is not ():
                    if cur_labels[i] is None:
                        new_labels.append(None)
                    else:
                        new_labels.append(list(newlabelarray))
            return Timeseries(new_array, new_tspan, new_labels)
        else:
            # advanced integer indexing
            is_fancy = tuple(not isinstance(x, _SliceType) for x in index)
            fancy_pos = tuple(i for i in range(len(index)) if is_fancy[i])
            nonfancy_pos = tuple(
                    i for i in range(len(index)) if not is_fancy[i])
            contiguous = (fancy_pos[-1] - fancy_pos[0] == len(fancy_pos) - 1)
            index = list(index)
            ix_arrays = [index[i] for i in fancy_pos]
            ix_arrays = np.broadcast_arrays(*ix_arrays)
            for j in range(len(fancy_pos)):
                if ix_arrays[j].shape is ():
                    ix_arrays[j] = np.expand_dims(ix_arrays[j], 0)
                index[fancy_pos[j]] = ix_arrays[j]
            index = tuple(index)
            ishape = index[fancy_pos[0]].shape # common shape all index arrays
            idim = len(ishape)
            assert(idim > 0)
            new_tspan = None
            if contiguous and not is_fancy[0]:
                new_tspan = self.tspan[index[0]]
                if (not isinstance(new_tspan, np.ndarray) or
                        new_tspan.shape is () or
                        new_tspan.shape[0] is 0):
                    # axis 0 has been sliced away: no longer a Timeseries
                    return np.asarray(new_array)
            # compute labels for the nonfancy output axes
            nonfancy_labels = []
            nonfancy_retained = []
            for i in nonfancy_pos:
                # make temp ndarray of labels to ensure correct slicing
                if cur_labels[i] is None:
                    labelarray = np.array([''] * cur_shape[i])
                else:
                    labelarray = np.array(cur_labels[i])
                newlabelarray = labelarray[index[i]]
                if newlabelarray.shape is ():
                    nonfancy_retained.append(False)
                else:
                    nonfancy_retained.append(True)
                    if cur_labels[i] is None or len(newlabelarray) is 0:
                        nonfancy_labels.append(None)
                    else:
                        nonfancy_labels.append(list(newlabelarray))
            # compute labels for the fancy output axes:
            #
            # For each fancy output axis k, call input axis i a 'candidate'
            # label source for k if k is the only nonconstant axis in the
            # indexing array for i.
            # We will give labels/tspan to output axis k from input axis i
            # only if i is the sole candidate label source for k.
            candidates = [[]] * idim
            # candidates[k] will be a list of candidate label sources for k
            for i in fancy_pos:
                nonconstant_ix_axes = []
                for k in range(idim):
                    n = ishape[k]
                    if n > 0:
                        partix = np.split(index[i], n, axis=k)
                        if not all(np.array_equal(
                                partix[0], partix[q]) for q in range(1, n)):
                            nonconstant_ix_axes.append(k)
                if len(nonconstant_ix_axes) is 1:
                    candidates[nonconstant_ix_axes[0]].append(i)
            fancy_labels = []
            for k in range(idim):
                if len(candidates[k]) is 1:
                    # then we can label this output axis
                    label_source = candidates[k][0]
                    if cur_labels[label_source] is None:
                        labelarray = np.array([''] * cur_shape[label_source])
                    else:
                        labelarray = np.array(cur_labels[label_source])
                    iix = [0] * idim
                    iix[k] = slice(None)
                    iix = tuple(iix)
                    newlabelarray = labelarray[index[label_source][iix]]
                    if newlabelarray.shape is not ():
                        if cur_labels[label_source] is None:
                            fancy_labels.append(None)
                        else:
                            fancy_labels.append(list(newlabelarray))
                    if k is 0 and (is_fancy[0] or not contiguous):
                        # then this output axis will be axis 0 of output
                        if label_source is 0:
                            new_tspan = self.tspan[index[0][iix]]
                            if (not isinstance(new_tspan, np.ndarray) or
                                    new_tspan.shape is () or
                                    new_tspan.shape[0] is 0):
                                # axis 0 has been sliced away: not a Timeseries
                                return np.asarray(new_array)
                            if not np.all(np.diff(new_tspan) > 0):
                                #tspan not monotonic increasing: not Timeseries
                                return np.asarray(new_array)
                        else:
                            #axis 0 no longer represents time: not a Timeseries
                            return np.asarray(new_array)
                else:
                    # not a 'sole candidate'
                    fancy_labels.append(None)
            if contiguous:
                # fancy output axes are put where the fancy input axes were:
                new_labels = []
                for i in range(0, fancy_pos[0]):
                    if nonfancy_retained.pop(0):
                        new_labels.append(nonfancy_labels.pop(0))
                new_labels.extend(fancy_labels)
                for i in range(fancy_pos[-1] + 1, cur_ndim):
                    if nonfancy_retained.pop(0):
                        new_labels.append(nonfancy_labels.pop(0))
            else:
                # not contiguous. fancy output axes move to the start:
                new_labels = fancy_labels + nonfancy_labels
            if new_tspan is None:
                return np.asarray(new_array)
            else:
                return Timeseries(new_array, new_tspan, new_labels)

    def __setitem__(self, index, value):
        #print('in __setitem__')
        return super(Timeseries, self).__setitem__(index, value)

    def __getslice__(self, i, j):
        #print('in __getslice__')
        return self.__getitem__(slice(i, j, None))

    def __setslice__(self, i, j, value):
        #print('in __setslice__')
        return self.__setitem__(slice(i, j, None), value)

    def __repr__(self):
        classname = self.__class__.__name__
        if self.tspan.shape is ():
            first = last = self.tspan
        else:
            first = self.tspan[0]
            last = self.tspan[-1]
        head = u'<%s of shape %s from time %f to %f>:\n' % (
            classname, self.shape, first, last)
        repr_tspan = 'tspan=' + repr(self.tspan)
        if len(repr_tspan) > 160:
            repr_tspan = 'tspan=array([ %f, ..., %f ])' % (first, last)
        content = repr(np.asarray(self)).replace('array', 
            classname, 1).rstrip(')') + ', \n' + repr_tspan
        if all(l is None for l in self.labels):
            labelsrep = ''
        else:
            labelsrep = ', \nlabels=['
            for lab in self.labels:
                if len(repr(lab)) > 160:
                    labelsrep += '[%s, ..., %s], ' % (lab[0], lab[-1])
                else:
                    labelsrep += repr(lab) + ', '
            labelsrep = labelsrep[:-2] + ']'
        return head + content + labelsrep + ')'

    def __reduce__(self):
        """Support pickling Timeseries instances by saving __dict__"""
        recon, initargs, state = super(Timeseries, self).__reduce__()
        tsstate = tuple((state, self.__dict__)) # nested, to avoid tuple copying
        return tuple((recon, initargs, tsstate))
        
    def __setstate__(self, tsstate):
        """Support unpickling Timeseries instances by loading __dict__"""
        super(Timeseries, self).__setstate__(tsstate[0])
        self.__dict__.update(tsstate[1])

    def __distob_scatter__(self, axis=-1, destination=None, blocksize=None):
        """Turn a Timeseries into a distributed timeseries"""
        import distob
        from nsim import DistTimeseries
        if axis is None:
            return distob.distob._scatter_ndarray(self, None, destination)
        if axis == 0:
            raise ValueError(u'Currently cannot distribute the time axis')
        if axis < 0:
            axis = self.ndim + axis
        dar = distob.distob._scatter_ndarray(self, axis,
                                             destination, blocksize)
        axlabels = self.labels[axis]
        return DistTimeseries([rts for rts in dar._subarrays], axis, axlabels)

    def absolute(self):
        """Calculate the absolute value element-wise.

        Returns:
          absolute (Timeseries):
            Absolute value. For complex input (a + b*j) gives sqrt(a**a + b**2)
        """
        return Timeseries(np.absolute(self), self.tspan, self.labels)

    def abs(self):
        """Calculate the absolute value element-wise."""
        return self.absolute()

    def angle(self, deg=False):
        """Return the angle of the complex argument.

        Args:
          deg (bool, optional):
            Return angle in degrees if True, radians if False (default).

        Returns:
          angle (Timeseries):
            The counterclockwise angle from the positive real axis on
            the complex plane, with dtype as numpy.float64.
        """
        if self.dtype.str[1] != 'c':
            warnings.warn('angle() is intended for complex-valued timeseries',
                          RuntimeWarning, 1)
        return Timeseries(np.angle(self, deg=deg), self.tspan, self.labels)

    # Some of the usual array operations on Timeseries return a plain ndarray. 
    # This depends on whether a time axis is present in the result:

    def flatten(self, order='C'):
        return np.asarray(self).flatten(order)

    def ravel(self, order='C'):
        return np.asarray(self).ravel(order)

    def swapaxes(self, axis1, axis2):
        """Interchange two axes of a Timeseries."""
        if self.ndim <=1 or axis1 == axis2:
            return self
        ar = np.asarray(self).swapaxes(axis1, axis2)
        if axis1 != 0 and axis2 != 0:
            # then axis 0 is unaffected by the swap
            labels = self.labels[:]
            labels[axis1], labels[axis2] = labels[axis2], labels[axis1]
            return Timeseries(ar, self.tspan, labels)
        return ar

    def transpose(self, *axes):
        """Permute the dimensions of a Timeseries."""
        if self.ndim <= 1:
            return self
        ar = np.asarray(self).transpose(*axes)
        if axes[0] != 0:
            # then axis 0 is unaffected by the transposition
            newlabels = [self.labels[ax] for ax in axes]
            return Timeseries(ar, self.tspan, newlabels)
        else:
            return ar

    def argmin(self, axis=None, out=None):
        return np.asarray(self).argmin(axis, out)

    def argmax(self, axis=None, out=None):
        return np.asarray(self).argmax(axis, out)

    def reshape(self, newshape, order='C'):
        """If axis 0 is unaffected by the reshape, then returns a Timeseries,
        otherwise returns an ndarray. Preserves labels of axis j only if all 
        axes<=j are unaffected by the reshape.  
        See ``numpy.ndarray.reshape()`` for more information
        """
        oldshape = self.shape
        ar = np.asarray(self).reshape(newshape, order=order)
        if (newshape is -1 and len(oldshape) is 1 or
                (isinstance(newshape, numbers.Integral) and 
                    newshape == oldshape[0]) or 
                (isinstance(newshape, Sequence) and
                    (newshape[0] == oldshape[0] or
                     (newshape[0] is -1 and np.array(oldshape[1:]).prod() ==
                                            np.array(newshape[1:]).prod())))):
            # then axis 0 is unaffected by the reshape
            newlabels = [None] * ar.ndim
            i = 1
            while i < ar.ndim and i < self.ndim and ar.shape[i] == oldshape[i]:
                newlabels[i] = self.labels[i]
                i += 1
            return Timeseries(ar, self.tspan, newlabels)
        else:
            return ar

    def min(self, axis=None, out=None):
        if (axis is 0 or 
                axis is None or
                self.ndim is 1 or 
                isinstance(axis, _TupleType) and 0 in axis):
            return np.asarray(self).min(axis, out)
        else:
            ar = super(Timeseries, self).min(axis, out)
            if isinstance(axis, numbers.Number):
                axis = (axis,)
            new_labels = []
            for i in range(self.ndim):
                if i not in axis:
                    new_labels.append(self.labels[i])
            return Timeseries(ar, self.tspan, new_labels)

    def max(self, axis=None, out=None):
        if (axis is 0 or 
                axis is None or
                self.ndim is 1 or 
                isinstance(axis, _TupleType) and 0 in axis):
            return np.asarray(self).max(axis, out)
        else:
            ar = super(Timeseries, self).max(axis, out)
            if isinstance(axis, numbers.Number):
                axis = (axis,)
            new_labels = []
            for i in range(self.ndim):
                if i not in axis:
                    new_labels.append(self.labels[i])
            return Timeseries(ar, self.tspan, new_labels)

    def ptp(self, axis=None, out=None):
        if (axis is 0 or 
                axis is None or
                self.ndim is 1 or 
                isinstance(axis, _TupleType) and 0 in axis):
            return np.asarray(self).ptp(axis, out)
        else:
            ar = super(Timeseries, self).ptp(axis, out)
            if isinstance(axis, numbers.Number):
                axis = (axis,)
            new_labels = []
            for i in range(self.ndim):
                if i not in axis:
                    new_labels.append(self.labels[i])
            return Timeseries(ar, self.tspan, new_labels)

    def mean(self, axis=None, dtype=None, out=None, keepdims=False):
        if axis == -1:
            axis = self.ndim
        if keepdims:
            if axis is None:
                out_shape = [1] * self.ndim
            else:
                out_shape = list(self.shape)
                out_shape[axis] = 1
        if (axis is 0 or 
                axis is None or
                self.ndim is 1 or 
                isinstance(axis, _TupleType) and 0 in axis):
            res = np.asarray(self).mean(axis, dtype, out)
            if keepdims:
                res = res.reshape(out_shape)
        else:
            ar = super(Timeseries, self).mean(axis, dtype, out)
            if isinstance(axis, numbers.Number):
                axis = (axis,)
            new_labels = []
            for i in range(self.ndim):
                if i not in axis:
                    new_labels.append(self.labels[i])
            res = Timeseries(ar, self.tspan, new_labels)
            if keepdims:
                res = res.reshape(out_shape)
        return res

    def std(self, axis=None, dtype=None, out=None, ddof=0):
        if (axis is 0 or 
                axis is None or
                self.ndim is 1 or 
                isinstance(axis, _TupleType) and 0 in axis):
            return np.asarray(self).std(axis, dtype, out, ddof)
        else:
            ar = super(Timeseries, self).std(axis, dtype, out)
            if isinstance(axis, numbers.Number):
                axis = (axis,)
            new_labels = []
            for i in range(self.ndim):
                if i not in axis:
                    new_labels.append(self.labels[i])
            return Timeseries(ar, self.tspan, new_labels)

    def var(self, axis=None, dtype=None, out=None, ddof=0):
        if (axis is 0 or 
                axis is None or
                self.ndim is 1 or 
                isinstance(axis, _TupleType) and 0 in axis):
            return np.asarray(self).var(axis, dtype, out, ddof)
        else:
            ar = super(Timeseries, self).var(axis, dtype, out)
            if isinstance(axis, numbers.Number):
                axis = (axis,)
            new_labels = []
            for i in range(self.ndim):
                if i not in axis:
                    new_labels.append(self.labels[i])
            return Timeseries(ar, self.tspan, new_labels)

    def merge(self, ts):
        """Merge another timeseries with this one
        Arguments:
          ts (Timeseries): The two timeseries being merged must have the
            same shape except for axis 0.
        Returns: 
          Resulting merged timeseries which can have duplicate time points.
        """
        if ts.shape[1:] != self.shape[1:]:
            raise ValueError('Timeseries to merge must have compatible shapes')
        indices = np.vstack((self.tspan, ts.tspan)).argsort()
        return np.vstack((self, ts))[indices]

    def expand_dims(self, axis):
        """Insert a new axis, at a given position in the array shape
        Args:
          axis (int): Position (amongst axes) where new axis is to be inserted.
        """
        if axis == -1:
            axis = self.ndim
        array = np.expand_dims(self, axis)
        if axis == 0:
            # prepended an axis: no longer a Timeseries
            return array
        else:
            new_labels = self.labels.insert(axis, None)
            return Timeseries(array, self.tspan, new_labels)

    def concatenate(self, tup, axis=0):
        """Join a sequence of Timeseries to this one
        Args: 
          tup (sequence of Timeseries): timeseries to be joined with this one.
            They must have the same shape as this Timeseries, except in the
            dimension corresponding to `axis`.
          axis (int, optional): The axis along which timeseries will be joined.
        Returns:
          res (Timeseries or ndarray)
        """
        if not isinstance(tup, Sequence):
            tup = (tup,)
        if tup is (None,) or len(tup) is 0:
            return self
        tup = (self,) + tuple(tup)
        new_array = np.concatenate(tup, axis)
        if not all(hasattr(ts, 'tspan') and 
                   hasattr(ts, 'labels') for ts in tup):
            return new_array
        if axis == 0:
            starts = [ts.tspan[0] for ts in tup]
            ends = [ts.tspan[-1] for ts in tup]
            if not all(starts[i] > ends[i-1] for i in range(1, len(starts))):
                # series being joined are not ordered in time. not Timeseries
                return new_array
            else:
                new_tspan = np.concatenate([ts.tspan for ts in tup])
        else:
            new_tspan = self.tspan
        new_labels = [None]
        for ax in range(1, new_array.ndim):
            if ax == axis:
                axislabels = []
                for ts in tup:
                    if ts.labels[axis] is None:
                        axislabels.extend('' * ts.shape[axis])
                    else:
                        axislabels.extend(ts.labels[axis])
                if all(lab == '' for lab in axislabels):
                    new_labels.append(None)
                else:
                    new_labels.append(axislabels)
            else:
                # non-concatenation axis
                axlabels = tup[0].labels[ax]
                if not all(ts.labels[ax] == axlabels for ts in tup[1:]):
                    # series to be joined do not agree on labels for this axis
                    axlabels = None
                new_labels.append(axlabels)
        return self.__new__(self.__class__, new_array, new_tspan, new_labels)

    def split(self, indices_or_sections, axis=0):
        """Split a timeseries into multiple sub-timeseries"""
        if not isinstance(indices_or_sections, numbers.Integral):
            raise Error('splitting by array of indices is not yet implemented')
        n = indices_or_sections
        if self.shape[axis] % n != 0:
            raise ValueError("Array split doesn't result in an equal division")
        step = self.shape[axis] / n
        pieces = []
        start = 0
        while start < self.shape[axis]:
            stop = start + step
            ix = [slice(None)] * self.ndim
            ix[axis] = slice(start, stop)
            ix = tuple(ix)
            pieces.append(self[ix])
            start += step
        return pieces


class _Timeslice(object):
    """Implements the `t` attribute of Timeseries that allows slicing by time
    instead of by array index. For example, timeseries.t[9.5:30:0.1] 
    resamples the time range from t=9.5 to t=30 in 0.1 second increments.
    """
    def __init__(self, ts):
        """Args: ts: Timeseries"""
        self.ts = ts

    def __getitem__(self, index):
        #print('in timeslice getitem, index type %s, index is %s' % (
        #    type(index), repr(index)))
        ts = self.ts
        dt = (ts.tspan[-1] - ts.tspan[0]) / (len(ts) - 1)
        if isinstance(index, numbers.Number):
            if not (index >= ts.tspan[0] and index <= ts.tspan[-1]):
                raise ValueError('time %s not in valid range %g to %g' % (
                        index, ts.tspan[0], ts.tspan[-1]))
            newix = ts.tspan.searchsorted(index)
            return ts[newix]
        elif isinstance(index, _SliceType):
            if index.start is None or index.start < ts.tspan[0]:
                startt = ts.tspan[0]
            else:
                startt = index.start
            if index.stop is None or index.stop > ts.tspan[-1]:
                stopt = ts.tspan[-1] + dt/2
            else:
                stopt = index.stop
            start, stop = ts.tspan.searchsorted([startt, stopt])
            if index.step is None or index.step <= dt:
                return ts[slice(start, stop, None)]
            else:
                times = np.arange(startt, stopt, index.step)
                if times[-1] == stopt:
                    times = times[:-1]
                indices = ts.tspan.searchsorted(times)
                if indices[-1] == len(ts.tspan):
                    indices = indices[:-1]
                return ts[indices]
        elif isinstance(index, _EllipsisType) or index is None:
            return ts[index]
        elif isinstance(index, np.ndarray) and index.ndim is 1:
            indices = ts.tspan.searchsorted(index)
            if indices[-1] == len(ts.tspan):
                indices = indices[:-1]
            return ts[indices]
        elif isinstance(index, _TupleType):
            timeix = index[0]
            otherix = index[1:]
            ts1 = ts.t[timeix]
            if ts1.ndim < ts.ndim:
                if len(otherix) is 1:
                    otherix = otherix[0]
                return ts1[otherix]
            else:
                return ts1[(slice(None),) + otherix]
        else:
            raise TypeError("Time slicing can't handle that type of index yet")

    def __setitem__(self, index, value):
        #TODO update logic to match __getitem__
        ts = self.ts
        dt = (ts.tspan[-1] - ts.tspan[0]) / (len(ts) - 1)
        if isinstance(index, numbers.Number):
            newix = ts.tspan.searchsorted(index)
            return ts.__setitem__(newix, value)
        elif isinstance(index, _SliceType):
            if index.step is None:
                start, stop = ts.tspan.searchsorted(index.start, index.stop)
                return ts.__setitem__(slice(start, stop, None), value)
            else:
                n = np.floor_divide(index.start - index.stop, index.step)
                times = np.linspace(index.start, index.stop, n, endpoint=False)
                indices = ts.tspan.searchsorted(times)
                if indices[-1] == len(ts.tspan):
                    indices = indices[:-1]
                return ts.__setitem__(indices, value)
        elif isinstance(index, _EllipsisType) or index is None:
            return ts.__setitem__(index, value)
        elif isinstance(index, np.ndarray) and index.ndim is 1:
            indices = ts.tspan.searchsorted(index)
            if indices[-1] == len(ts.tspan):
                indices = indices[:-1]
            return ts.__setitem__(indices, value)
        elif isinstance(index, _TupleType):
            timeix = index[0]
            ts = ts.t[timeix]
            otherix = index[1:]
            return ts.__setitem__(otherix, value)
        else:
            raise TypeError("Time slicing can't handle that type of index yet")

    def __getslice__(self, i, j):
        return self.__getitem__(slice(i, j, None))

    def __setslice__(self, i, j, value):
        return self.__setitem__(slice(i, j, None), value)

    def __repr__(self):
        fs = (1.0*len(self.ts) - 1) / (self.ts.tspan[-1] - self.ts.tspan[0])
        s = u'Time range %g to %g with average sample rate %g Hz\n' % (
                self.ts.tspan[0], self.ts.tspan[-1], fs)
        return s


def merge(tup):
    """Merge several timeseries
    Arguments:
      tup: sequence of Timeseries, with the same shape except for axis 0
    Returns: 
      Resulting merged timeseries which can have duplicate time points.
    """
    if not all(tuple(ts.shape[1:] == tup[0].shape[1:] for ts in tup[1:])):
        raise ValueError('Timeseries to merge must have compatible shapes')
    indices = np.vstack(tuple(ts.tspan for ts in tup)).argsort()
    return np.vstack((tup))[indices]



"""Example of using a model defined with a single complex variable.

dz = (-z + j (delta + g cos(omega t)) z + j epsilon) dt + sigma exp(4jt) dW

With parameters below the system has noise driven oscillations around a stable
fixed point.

Using complex vector variables is ok, too. (numpy arrays of complex128)
nsim will look at y0 to determine what type of variable is being used.
"""

import nsim
import numpy as np
import matplotlib.pyplot as plt


class Osc(nsim.StratonovichModel):
    delta = 2.0
    epsilon = 100.0
    sigma = 10.0
    g = 1.0
    omega = 2.0
    y0 = 0.0 + 0.0j

    def f(self, y, t):
        return (-y + 1j*(self.delta + self.g*np.cos(self.omega*t))*y +
                1j*self.epsilon)

    def G(self, y, t):
        return self.sigma * np.exp(4j*t)
    

sims = nsim.RepeatedSim(Osc, T=1440.0, repeat=10)
ts = sims.timeseries
means = ts.mean(axis=0)

phases = (ts - means).angle()

phases.plot(title='phase at each node')
phases[:,:,3].t[100:160].plot(title='phase') # show 60 seconds of node 3

print('mean period is %g seconds' % phases.periods().mean())

r = (ts - means).abs()
r.plot(title='amplitude at each node')

# show distribution of mean reversion times
rtimes = r.first_return_times(r.mean())
plt.figure()
plt.hist(rtimes, bins=150, range=(0.05, 2.9))
plt.title('Distribution of mean reversion times for amplitude')
plt.show(block=False)

print('\nmean reversion time is %g seconds' % rtimes.mean())
